"""
Computes categorical metrics given a crosstab df.
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Iterable, Optional, Union, Dict
from numbers import Number

# deprecation warnings: DeprecationWarning: In a future version, `df.iloc[:, i] = newvals` will attempt to set the values inplace instead of always setting a new array. To retain the old behavior, use either `df[df.columns[i]] = newvals` or, if columns are non-unique, `df.isetitem(i, newvals)
import warnings

import numpy as np
import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame

from gval import CatStats
from gval.utils.schemas import (
    Crosstab_df,
    Sample_identifiers,
    Subsample_identifiers,
    Metrics_df,
)


def _create_set_dict(categories):
    """Creates set valued dictionary"""
    if not isinstance(categories, dict):
        categories = {"candidate": categories, "benchmark": categories}
    else:
        if "candidate" not in categories or "benchmark" not in categories:
            raise ValueError(
                "Make sure both candidate and benchmark entries are in the dictionary passed"
                "in to negative_categories, positive_categories, or both."
            )

    categories["candidate"] = _convert_to_set(categories["candidate"])
    categories["benchmark"] = _convert_to_set(categories["benchmark"])

    return categories


def _convert_to_set(data):
    """Converts data to set"""
    if data is None:
        data = set()
    else:
        data = {data} if isinstance(data, Number) else set(data)

    return data


def _get_unique_values(crosstab_df, column):
    """Get unique values from a Dataframe"""

    return set(
        [
            item
            for item in crosstab_df.loc[:, column].to_numpy().ravel()
            if not isinstance(item, list)
        ]
    )


def _handle_positive_negative_categories(
    crosstab_df, positive_categories, negative_categories
):  # pragma: no cover
    """Input handling for use case with positive and negative categories"""

    positive_categories = _create_set_dict(positive_categories)
    negative_categories = _create_set_dict(negative_categories)

    # input check to make sure the same value isn't repeated in positive and negative categories
    if positive_categories["candidate"].intersection(negative_categories["candidate"]):
        raise ValueError(
            "Value is shared in candidate positive and negative categories."
        )

    if positive_categories["benchmark"].intersection(negative_categories["benchmark"]):
        raise ValueError(
            "Value is shared in benchmark positive and negative categories."
        )

    # Finds the unique values in the sample's candidate and benchmark values
    candidate_unique_values = _get_unique_values(crosstab_df, "candidate_values")
    benchmark_unique_values = _get_unique_values(crosstab_df, "benchmark_values")

    # Share unique values if both candidate and benchmark share the same categories
    if (
        positive_categories["candidate"] == positive_categories["benchmark"]
        and negative_categories["candidate"] == negative_categories["benchmark"]
    ):
        candidate_unique_values = (
            benchmark_unique_values
        ) = candidate_unique_values.union(benchmark_unique_values)

    # This checks that user passed positive or negative categories exist in sample df
    def categories_exist_in_crosstab_df(categories, cat_name, unique_values):
        for c in categories:
            if c not in unique_values:
                raise ValueError(f"{cat_name} category {c} not found in crosstab df.")

    categories_exist_in_crosstab_df(
        positive_categories["candidate"], "positive", candidate_unique_values
    )
    categories_exist_in_crosstab_df(
        negative_categories["candidate"], "negative", candidate_unique_values
    )
    categories_exist_in_crosstab_df(
        positive_categories["benchmark"], "positive", benchmark_unique_values
    )
    categories_exist_in_crosstab_df(
        negative_categories["benchmark"], "negative", benchmark_unique_values
    )

    return positive_categories, negative_categories


@pa.check_types
def _compute_categorical_metrics(
    crosstab_df: DataFrame[Crosstab_df],
    positive_categories: Optional[
        Union[Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]]
    ],
    negative_categories: Optional[
        Union[Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]]
    ] = None,
    metrics: Union[str, Iterable[str]] = "all",
    average: str = "micro",
    weights: Optional[Iterable[Number]] = None,
    sampling_average: Optional[str] = None,
) -> DataFrame[Metrics_df]:
    """
    Computes categorical metrics from a crosstab df.

    Parameters
    ----------
    crosstab_df : DataFrame[Crosstab_df]
        Crosstab DataFrame with candidate, benchmark, and agreement values as well as the counts for each occurrence.
    positive_categories : Optional[Union[Number, Iterable[Number]]]
        Number or list of numbers representing the values to consider as the positive condition. For average types "macro" and "weighted", this represents the categories to compute metrics for.
    metrics : Union[str, Iterable[str]], default = "all"
        String or list of strings representing metrics to compute.
    negative_categories : Optional[Union[Number, Iterable[Number]]], default = None
        Number or list of numbers representing the values to consider as the negative condition. This should be set to None when no negative categories are used or when the average type is "macro" or "weighted".
    average : str, default = "micro"
        Type of average to use when computing metrics. Options are "micro", "macro", and "weighted".
        Micro weighing computes the conditions, tp, tn, fp, and fn, for each category and then sums them.
        Macro weighing computes the metrics for each category then averages them.
        Weighted average computes the metrics for each category then averages them weighted by the number of weights argument in each category.
    weights : Optional[Iterable[Number]], default = None
        Weights to use when computing weighted average. Elements correspond to positive categories in order.

        Example:

        `positive_categories = [1, 2]; weights = [0.25, 0.75]`
    sampling_average: Optional[str], default = None
        Way to aggregate statistics for subsamples if provided. Options are "sample", "band", and "full-detail"
        Sample calculates metrics and averages the results by subsample
        Band calculates metrics and averages all the metrics by band
        Full-detail does not aggregation on subsample or band

    Returns
    -------
    DataFrame[Metrics_df]
        Metrics DF with computed metrics per sample.

    Raises
    ------
    ValueError
        Value is shared in positive and negative categories.
    ValueError
        Category not found in crosstab df.
    ValueError
        Cannot use average type with only one positive category.
    ValueError
        Number of weights must be the same as the number of positive categories.
    ValueError
        Cannot use average type with negative_categories as not None. Set negative_categories to None for this average type.

    References
    ----------
    .. [1] `Evaluation of binary classifiers <https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers>`_
    .. [2] `7th International Verification Methods Workshop <https://www.cawcr.gov.au/projects/verification/#Contingency_table>`_
    .. [3] `3.3. Metrics and scoring: quantifying the quality of predictions <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
    """

    #########################################################################################
    # input handling

    # make copy to avoid modifying original
    crosstab_df = crosstab_df.copy()

    positive_categories, negative_categories = _handle_positive_negative_categories(
        crosstab_df, positive_categories, negative_categories
    )

    # check to make sure that macro and weighted average are not used with only one category
    if average in ["macro", "weighted"] and len(positive_categories["candidate"]) == 1:
        raise ValueError(
            f"Cannot use average type '{average}' with only one positive category."
        )

    # check to make sure length of weights is the same as positive categories
    if average == "weighted":
        if len(weights) != len(positive_categories["candidate"]):
            raise ValueError(
                f"Number of positive category weights ({len(weights)}) must be the same as the number of positive "
                f"categories ({len(positive_categories['candidate'])})."
            )

    # make sure negative categories are not used with macro or weighted average
    if average in ["macro", "weighted"] and negative_categories["candidate"]:
        raise ValueError(
            f"Cannot use average type '{average}' with negative_categories as not None. "
            f"Set negative_categories to None for this average type."
        )

    if (
        average in ["macro", "weighted"]
        and positive_categories["candidate"] != positive_categories["benchmark"]
    ):
        raise ValueError(
            f"Cannot use average type '{average}' with different classes in positive candidate and benchmark values. "
            f"Make sure they are the same."
        )

    #########################################################################################
    # assign conditions

    def assign_condition_to_pairing(row, positive_categories, negative_categories):
        """This is used to create a conditions column"""
        # predicted positive
        if row["candidate_values"] in positive_categories["candidate"]:
            if row["benchmark_values"] in positive_categories["benchmark"]:
                return "tp"
            elif (
                row["benchmark_values"] in negative_categories["benchmark"]
                or len(negative_categories["benchmark"]) == 0
            ):
                return "fp"

        # predicted negative
        elif (
            row["candidate_values"] in negative_categories["candidate"]
            or len(negative_categories["candidate"]) == 0
        ):
            if row["benchmark_values"] in positive_categories["benchmark"]:
                return "fn"
            elif row["benchmark_values"] in negative_categories["benchmark"]:
                return "tn"

    #########################################################################################
    # compute metrics

    def compute_metrics_per_sample(sample):
        """Computes the metric values on a per sample basis"""

        # this turns the sample df to a groupby object on conditions
        conditions_series = (
            sample.loc[:, ["counts", "conditions"]]
            .groupby("conditions")
            .sum()[
                "counts"
            ]  # this is only necessary for micro average but left in place regardless
        )
        # .loc[Conditions_df.columns()] # used for ordering conditions

        # compute metric values and return names
        metric_values, metric_names = CatStats.process_statistics(
            metrics, **conditions_series.to_dict()
        )

        metrics_series = pd.Series(metric_values, index=metric_names)

        return pd.concat([conditions_series, metrics_series])

    #########################################################################################
    # computes metrics per category or micro level category

    def compute_metrics_per_category(pos_cats, neg_cats, groupby_cols):
        # assign conditions to crosstab_df
        crosstab_df["conditions"] = crosstab_df.apply(
            assign_condition_to_pairing, axis=1, args=(pos_cats, neg_cats)
        )

        # groupby sample identifiers then compute metrics
        metric_df = (
            crosstab_df.groupby(groupby_cols)
            .apply(compute_metrics_per_sample)
            .reset_index()
        )

        if "tn" not in metric_df.columns:
            metric_df.insert(
                loc=metric_df.columns.get_loc("tp"), column="tn", value=np.nan
            )

        # warnings for deprecation
        with warnings.catch_warnings(record=True):
            # Filter warnings by category, in this case, DeprecationWarning
            warnings.simplefilter("always", DeprecationWarning)
            if len(pos_cats["candidate"]) == 1:
                metric_df.insert(
                    loc=metric_df.columns.get_loc("fn"),
                    column="positive_categories",
                    value=None,
                )
                metric_df.loc[:, "positive_categories"] = pos_cats["candidate"].pop()
            elif len(metric_df) == len(pos_cats["candidate"]):
                metric_df.insert(
                    loc=metric_df.columns.get_loc("fn"),
                    column="positive_categories",
                    value=None,
                )
                metric_df.loc[:, "positive_categories"] = list(pos_cats)

        return metric_df

    #########################################################################################
    # Get subsampling_grouping
    if sampling_average == "band":
        groupby = Subsample_identifiers.columns()
    elif sampling_average == "subsample" or sampling_average is None:
        groupby = Sample_identifiers.columns()
    else:
        groupby = list(
            np.ravel([Sample_identifiers.columns(), Subsample_identifiers.columns()])
        )  # full-detail

    # replaces metrics variable with all metrics if metrics is set to all
    metrics = CatStats.available_functions() if metrics == "all" else metrics

    # actually compute metrics given averaging scheme
    if average == "micro":
        metric_df = compute_metrics_per_category(
            positive_categories, negative_categories, groupby
        )
    elif (average == "macro") or (average == "weighted") or (average is None):
        # Make copies to mutate
        positive_copy = positive_categories.copy()
        negative_copy = negative_categories.copy()
        metrics_dfs = []

        # One vs All Class Multi-categorical
        for pos_cat in positive_categories["candidate"]:
            positive_copy["candidate"] = positive_copy["benchmark"] = {pos_cat}
            negative_copy["candidate"] = negative_copy[
                "benchmark"
            ] = positive_categories["candidate"].difference({pos_cat})
            metrics_dfs.append(
                compute_metrics_per_category(positive_copy, negative_copy, groupby)
            )

        metric_df = pd.concat(metrics_dfs)

    #########################################################################################
    # aggregate metrics

    if average is None:
        metric_df.reset_index(drop=True, inplace=True)

    elif average == "micro":
        metric_df = metric_df.drop(
            columns=["positive_categories"], errors="ignore"
        ).reset_index(drop=True)

    elif average == "macro":
        metric_df = (
            metric_df.groupby(groupby)
            .mean(numeric_only=True)
            .drop(
                columns=["fn", "fp", "tn", "tp", "positive_categories"], errors="ignore"
            )
            .reset_index()
        )

    elif average == "weighted":
        if (weights is not None) & (average == "weighted"):
            # turn weights into a dict paired with positive categories
            weights = dict(zip(metric_df["positive_categories"], weights))

            def assign_weights(row):
                """This is used to create a weights column"""
                try:
                    return weights[row["positive_categories"]]
                except KeyError:
                    return np.nan

            # add a column in crosstab_df called weights
            metric_df.insert(
                loc=metric_df.columns.get_loc("positive_categories"),
                column="weights",
                value=np.nan,
            )

            metric_df["weights"] = metric_df.apply(assign_weights, axis=1)

        # TODO: WHERE ARE THE INDEXES NOT BEING RESET PREVIOUSLY?
        metric_df.reset_index(drop=True, inplace=True)

        # compute weighted average
        weighted_metrics = (
            metric_df.loc[:, metrics]
            .multiply(metric_df.loc[:, "weights"], axis=0)
            .reset_index(drop=True)
        )

        # add weighted metrics to metric_df
        metric_df.loc[:, metrics] = weighted_metrics

        # take average of weighted metrics
        metric_df = (
            metric_df.groupby(Sample_identifiers.columns())
            .sum(numeric_only=True)
            .drop(
                columns=["fn", "fp", "tn", "tp", "weights", "positive_categories"],
                errors="ignore",
            )
            .divide(metric_df.loc[:, "weights"].sum())
            .reset_index()
        )

    if sampling_average == "band":
        metric_df.insert(1, "band", "averaged")
    elif sampling_average == "subsample":
        metric_df.insert(0, "subsample", "averaged")

    return metric_df



"""
Comparison functionality
 - Includes pairing and comparison functions
 - crosstabbing functionality
 - Would np.meshgrid lend itself to pairing problem?
TODO:
    - Have not tested parallel case.
    - How to handle xr.Datasets, multiple bands, and multiple variables.
    - consider making a function registry to store pairing functions and their names in a dictionary
        - [Guide to function registration in Python with decorators](https://blog.miguelgrinberg.com/post/the-ultimate-guide-to-python-decorators-part-i-function-registration)
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Iterable, Tuple, Optional, Dict
from numbers import Number

import numpy as np
from numba import vectorize, uint8, int32, int64, float32, float64, boolean


# Numba Type Definitions
one_param_function_types = [
    uint8(uint8),
    int32(int32),
    int64(int64),
    float32(float32),
    float64(float64),
]
two_param_function_types = [
    uint8(uint8, uint8),
    int32(int32, int32),
    int64(int64, int64),
    float32(float32, float32),
    float64(float64, float64),
]
not_natural_number_types = [
    boolean(uint8, boolean),
    boolean(int32, boolean),
    int64(int64, boolean),
    float32(float32, boolean),
    float64(float64, boolean),
]


@vectorize(not_natural_number_types, nopython=True)
def _is_not_natural_number(
    x: Number, raise_exception: bool
) -> bool:  # pragma: no cover
    """
    Checks value to see if it is a natural number or two non-negative integer [0, 1, 2, 3, 4, ...)

    Parameters
    ----------
    x : Number
        Number to test.
    raise_exception : bool
        Raise exception or return bool.

    Returns
    -------
    bool
        Is natural number or not. np.nan treated as natural.

    Raises
    ------
    ValueError
        If x not a natural number.
    """
    # checks to make sure it's not a nan value
    if np.isnan(x):
        return False  # treated as natural for this use case

    # checks for non-negative and whole number
    elif (x < 0) | ((x - int64(x)) != 0):
        if raise_exception:
            raise ValueError(
                "Non natural number found (non-negative integers, excluding Inf) [0, 1, 2, 3, 4, ...)"
            )
        else:
            return True

    # is natural
    else:
        return False


@vectorize(two_param_function_types, nopython=True)
def cantor_pair(c: Number, b: Number) -> Number:  # pragma: no cover
    """
    Produces unique natural number for two non-negative natural numbers (0,1,2,...)

    Parameters
    ----------
    c : Number
        Candidate map value.
    b : Number
        Benchmark map value.

    Returns
    -------
    Number
        Agreement map value.

    References
    ----------
    .. [1] `Cantor and Szudzik Pairing Functions <https://www.vertexfragment.com/ramblings/cantor-szudzik-pairing-functions/#signed-szudzik>`_
    """
    _is_not_natural_number(c, True)
    _is_not_natural_number(b, True)

    if np.isnan(c) or np.isnan(b):
        return np.nan
    else:
        return 0.5 * (c**2 + c + 2 * c * b + 3 * b + b**2)


@vectorize(two_param_function_types, nopython=True)
def szudzik_pair(c: Number, b: Number) -> Number:  # pragma: no cover
    """
    Produces unique natural number for two non-negative natural numbers (0,1,2,3,...).

    Parameters
    ----------
    c : Number
        Candidate map value.
    b : Number
        Benchmark map value.

    Returns
    -------
    Number
        Agreement map value.

    References
    ----------
    .. [1] `Cantor and Szudzik Pairing Functions <https://www.vertexfragment.com/ramblings/cantor-szudzik-pairing-functions/#signed-szudzik>`_
    """
    _is_not_natural_number(c, True)
    _is_not_natural_number(b, True)

    if np.isnan(c) or np.isnan(b):
        return np.nan
    else:
        return c**2 + c + b if c >= b else b**2 + c


@vectorize(one_param_function_types, nopython=True)
def _negative_value_transformation(x: Number) -> Number:  # pragma: no cover
    """
    Transforms negative values for use with pairing functions that only accept non-negative integers.

    Parameters
    ----------
    x : Number
        Negative number to be transformed.

    Returns
    -------
    Number
        Transformed value.

    References
    ----------
    .. [1] `Cantor and Szudzik Pairing Functions <https://www.vertexfragment.com/ramblings/cantor-szudzik-pairing-functions/#signed-szudzik>`_
    """
    if np.isnan(x):
        return x
    else:
        return 2 * x if x >= 0 else -2 * x - 1


@vectorize(two_param_function_types, nopython=True)
def cantor_pair_signed(c: Number, b: Number) -> Number:  # pragma: no cover
    """
    Output unique natural number for each unique combination of whole numbers using Cantor signed method.

    Parameters
    ----------
    c : Number
        Candidate map value.
    b : Number
        Benchmark map value.

    Returns
    -------
    Number
        Agreement map value.

    References
    ----------
    .. [1] `Cantor and Szudzik Pairing Functions <https://www.vertexfragment.com/ramblings/cantor-szudzik-pairing-functions/#signed-szudzik>`_
    """
    ct = _negative_value_transformation(c)
    bt = _negative_value_transformation(b)

    if np.isnan(c) or np.isnan(b):
        return np.nan
    else:
        return cantor_pair(ct, bt)


# from typing import TypeVar
#
# T = TypeVar("T")


@vectorize(two_param_function_types, nopython=True)
def szudzik_pair_signed(c: Number, b: Number) -> Number:  # pragma: no cover
    """
    Output unique natural number for each unique combination of whole numbers using Szudzik signed method._summary_

    Parameters
    ----------
    c : Number
        Candidate map value.
    b : Number
        Benchmark map value.

    Returns
    -------
    Number
        Agreement map value.

    References
    ----------
    .. [1] `Cantor and Szudzik Pairing Functions <https://www.vertexfragment.com/ramblings/cantor-szudzik-pairing-functions/#signed-szudzik>`_
    """
    ct = _negative_value_transformation(c)
    bt = _negative_value_transformation(b)

    if np.isnan(c) or np.isnan(b):
        return np.nan
    else:
        return szudzik_pair(ct, bt)


class PairingDict(dict):
    """
    Pairing dictionary class that enables to replace np.nans with a different value as np.nan != np.nan.

    Parameters
    ----------
    dict : dict
        Parent class.

    Attributes
    -------
    replacement_value
        Value to use instead of np.nan.
    """

    def __init__(self, *args, **kwargs):
        # make a tmp dict with
        tmp_dict = dict(*args, **kwargs)
        dtype = type(np.array(list(tmp_dict.keys()))[0][0])

        if issubclass(dtype, np.integer):
            self.replacement_value = np.iinfo(dtype).max
        else:
            self.replacement_value = np.finfo(dtype).max

        # initialize self
        super().__init__()

        # set the items
        for k, v in tmp_dict.items():
            self.__setitem__(k, v)

    def _replace_nans(self, key):
        """Replaces NaNs"""
        new_key = [None] * len(key)
        for i, k in enumerate(key):
            if k == self.replacement_value:
                new_key[i] = k
            elif np.isnan(k):  # cannot be a string
                new_key[i] = self.replacement_value
            else:
                new_key[i] = k

        return tuple(new_key)

    def __getitem__(self, key):
        key = self._replace_nans(key)
        # FIXME:  Returns nans when present in key but any other key in the mean time
        if self.replacement_value in key:
            return np.nan
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        key = self._replace_nans(key)
        super().__setitem__(key, value)

    def __contains__(self, key):  # pragma: no cover
        # FIXME: Why are these two lines not covered? pragma: no cover for the time being
        key = self._replace_nans(key)
        return super().__contains__(key)


def _make_pairing_dict(
    unique_candidate_values: Iterable, unique_benchmark_values: Iterable
) -> Dict[Tuple[Number, Number], Number]:
    """
    Creates a dict pairing each unique value in candidate and benchmark arrays.

    Parameters
    ----------
    unique_candidate_values : Iterable
        Unique values in candidate map to create pairing dict with.
    unique_benchmark_values : Iterable
        Unique values in benchmark map to create pairing dict with.
    Returns
    -------
    Dict[Tuple[Number, Number], Number]
        Dictionary with keys consisting of unique pairings of candidate and benchmark values with value of agreement map for given pairing.
    """
    from itertools import product

    # TODO: consider allow use of unique to acquire all values from candidate and benchmarks
    if (unique_candidate_values is None) | (unique_benchmark_values is None):
        raise ValueError(
            "When comparison_function argument is set to 'pairing_dict', must pass values for "
            "allow_candidate_values and allow_benchmark_values arguments."
        )

    pairing_dict = {
        k: v
        for v, k in enumerate(product(unique_candidate_values, unique_benchmark_values))
    }

    # return pairing_dict
    return PairingDict(pairing_dict)


def _make_pairing_dict_fn(
    pairing_dict: Optional[dict] = None,
    unique_candidate_values: Optional[Iterable] = None,
    unique_benchmark_values: Optional[Iterable] = None,
) -> Dict[Tuple[Number, Number], Number]:
    """
    Creates a unique value in candidate and benchmark arrays.

    Parameters
    ----------
    pairing_dict : Optional[dict], default = None
        User provided dictionary of unique encodings
    unique_candidate_values : Optional[Iterable], default = None
        Unique values in candidate map to create pairing dict with.
    unique_benchmark_values : Optional[Iterable], default = None
        Unique values in benchmark map to create pairing dict with.

    Returns
    -------
    Dict[Tuple[Number, Number], Number]
        Dictionary with keys consisting of unique pairings of candidate and benchmark values with value of agreement
        map for given pairing.
    """

    pairing_dict = (
        _make_pairing_dict(unique_candidate_values, unique_benchmark_values)
        if pairing_dict is None
        else PairingDict(pairing_dict)
    )

    # Populate
    replacement_value = pairing_dict.replacement_value
    keys = np.array(list(pairing_dict.keys()))
    keys1, keys2 = keys[:, 0], keys[:, 1]
    values = np.array(list(pairing_dict.values()))

    def pairing_dict_fn(
        c: Number,
        b: Number,
    ) -> Number:  # pragma: no cover
        """
        Produces a pairing dictionary that produces a unique result for every combination ranging from 256 to the number of combinations.

        Parameters
        ----------
        c : Number
            Candidate map value.
        b : Number
            Benchmark map value.

        Returns
        -------
        Number
            Agreement map value.

        Raises
        ------
        KeyError
            Value combination found not accounted for in pairing dictionary
        """

        nans = 0
        if np.isnan(c):
            c = replacement_value
            nans = 1

        if np.isnan(b):
            b = replacement_value
            nans = 1

        for key1, key2, value in zip(keys1, keys2, values):
            if c == key1 and b == key2:
                return value

        if nans == 1:
            return np.nan
        else:
            raise KeyError(
                "Value combination found not accounted for in pairing dictionary"
            )

    return vectorize(two_param_function_types, nopython=True)(pairing_dict_fn)


@vectorize(two_param_function_types, nopython=True)
def difference(c: Number, b: Number) -> Number:  # pragma: no cover
    """
    Calculates the difference between candidate and benchmark.

    Parameters
    ----------
    c : Number
        Candidate map value.
    b : Number
        Benchmark map value.

    Returns
    -------
    Number
        Difference between candidate and benchmark map values.
    """
    return c - b



"""
Test functionality for continuous value metrics.
"""

# __all__ = ['*']

from functools import wraps

import numpy as np
from pytest_cases import parametrize_with_cases

from tests.conftest import _assert_pairing_dict_equal
from gval.statistics import continuous_stat_funcs


def assert_logic_decorator(test_function):
    @wraps(test_function)
    def wrapper(candidate_map, benchmark_map, expected_value):
        # Call the test function
        result = test_function(candidate_map, benchmark_map, expected_value)

        # Implement the assert logic
        if isinstance(expected_value, dict):
            _assert_pairing_dict_equal(result, expected_value)
        else:
            np.testing.assert_almost_equal(result, expected_value)

    return wrapper


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="mean_absolute_error",
)
@assert_logic_decorator
def test_mean_absolute_error(candidate_map, benchmark_map, expected_value):
    return continuous_stat_funcs.mean_absolute_error(candidate_map - benchmark_map)


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="mean_squared_error",
)
@assert_logic_decorator
def test_mean_squared_error(candidate_map, benchmark_map, expected_value):
    return continuous_stat_funcs.mean_squared_error(candidate_map - benchmark_map)


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="root_mean_squared_error",
)
@assert_logic_decorator
def test_root_mean_squared_error(candidate_map, benchmark_map, expected_value):
    return continuous_stat_funcs.root_mean_squared_error(candidate_map - benchmark_map)


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="mean_signed_error",
)
@assert_logic_decorator
def test_mean_signed_error(candidate_map, benchmark_map, expected_value):
    return continuous_stat_funcs.mean_signed_error(candidate_map - benchmark_map)


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="mean_percentage_error",
)
@assert_logic_decorator
def test_mean_percentage_error(candidate_map, benchmark_map, expected_value):
    return continuous_stat_funcs.mean_percentage_error(
        candidate_map - benchmark_map, benchmark_map
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="mean_absolute_percentage_error",
)
@assert_logic_decorator
def test_mean_absolute_percentage_error(candidate_map, benchmark_map, expected_value):
    return continuous_stat_funcs.mean_absolute_percentage_error(
        candidate_map - benchmark_map, benchmark_map
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="mean_normalized_root_mean_squared_error",
)
@assert_logic_decorator
def test_mean_normalized_root_mean_squared_error(
    candidate_map, benchmark_map, expected_value
):
    return continuous_stat_funcs.mean_normalized_root_mean_squared_error(
        candidate_map - benchmark_map, benchmark_map
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="range_normalized_root_mean_squared_error",
)
@assert_logic_decorator
def test_range_normalized_root_mean_squared_error(
    candidate_map, benchmark_map, expected_value
):
    return continuous_stat_funcs.range_normalized_root_mean_squared_error(
        candidate_map - benchmark_map, benchmark_map
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="mean_normalized_mean_absolute_error",
)
@assert_logic_decorator
def test_mean_normalized_mean_absolute_error(
    candidate_map, benchmark_map, expected_value
):
    return continuous_stat_funcs.mean_normalized_mean_absolute_error(
        candidate_map - benchmark_map, benchmark_map
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="range_normalized_mean_absolute_error",
)
@assert_logic_decorator
def test_range_normalized_mean_absolute_error(
    candidate_map, benchmark_map, expected_value
):
    return continuous_stat_funcs.range_normalized_mean_absolute_error(
        candidate_map - benchmark_map, benchmark_map
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="coefficient_of_determination",
)
@assert_logic_decorator
def test_coefficient_of_determination(candidate_map, benchmark_map, expected_value):
    return continuous_stat_funcs.coefficient_of_determination(
        candidate_map - benchmark_map, benchmark_map
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_value",
    glob="symmetric_mean_absolute_percentage_error",
)
@assert_logic_decorator
def test_symmetric_mean_absolute_percentage_error(
    candidate_map, benchmark_map, expected_value
):
    return continuous_stat_funcs.symmetric_mean_absolute_percentage_error(
        candidate_map - benchmark_map, candidate_map, benchmark_map
    )



"""Functions to load or create datasets"""
from __future__ import annotations

__author__ = "Fernando Aristizabal"


from typing import Union, Optional, Tuple, Iterable
from numbers import Number
import ast
from collections import Counter

import pandas as pd
import rioxarray as rxr
import xarray as xr
import numpy as np
from tempfile import NamedTemporaryFile
from rio_cogeo.cogeo import cog_translate
from rio_cogeo.profiles import cog_profiles
from pystac.item_collection import ItemCollection

_MEMORY_STRATEGY = "normal"


def adjust_memory_strategy(strategy: str):
    """
    Tells GVAL how to address handling memory.  There are three modes currently available:

    normal:  Keeps all of xarray files in memory as usual
    moderate:  Either creates cloud optimized geotiffs and stores as temporary files and reloads or reloads file
    to be in lazily loaded stated
    aggressive: Does the same as moderate except loads with no cache so everything is read from disk

    There are tradeoffs with performance for choosing a strategy that conserves memory, adjust only as needed.

    Parameters
    ----------
    strategy : str, {'normal', 'moderate', 'aggressive'}
        Method to conserve memory

    Raises
    ------
    ValueError

    """

    if strategy in {"normal", "moderate", "aggressive"}:
        global _MEMORY_STRATEGY
        _MEMORY_STRATEGY = strategy

    else:
        raise ValueError(
            "Please select one of the following options for a memory strategy "
            "'normal', 'moderate', 'aggressive'"
        )


def get_current_memory_strategy() -> str:
    """
    Gets the current memory_strategy

    Returns
    -------
    str
        Memory optimization strategy
    """
    return _MEMORY_STRATEGY


def _handle_xarray_memory(
    data_obj: Union[xr.Dataset, xr.DataArray], make_temp: bool = False
) -> Union[xr.Dataset, xr.DataArray]:
    """
    Executes memory strategy plan depending on the set memory strategy.

    Nothing happens in this method when the plan is normal, only if the strategy is moderate or aggressive does
    any memory optimization occur.  If the parameter make_temp is true a temporary file with a lzw compressed
    cloud optimized geotiff will be made, otherwise the original file will be loaded.

    Parameters
    ----------
    data_obj : Union[xr.Dataset, xr.DataArray]
        Xarray object to handle_memory
    make_temp: bool
        Store data in a temporary file if in memory

    Returns
    -------
    Union[xr.Dataset, xr.DataArray]
        Memory handled xarray object
    """

    # Check to see if it is already dask and what memory strategy is in place
    band_as_var = True if isinstance(data_obj, xr.Dataset) else False
    cache = True if _MEMORY_STRATEGY != "aggressive" else False

    if _MEMORY_STRATEGY == "normal" or _check_dask_array(data_obj):  # do nothing
        return data_obj
    elif (
        make_temp is False
    ):  # reload the file with cache on or off depending on strategy
        file_name = (
            data_obj["band_1"].encoding["source"]
            if band_as_var
            else data_obj.encoding["source"]
        )
        new_obj = _parse_string_attributes(
            rxr.open_rasterio(
                file_name,
                mask_and_scale=True,
                band_as_variable=band_as_var,
                cache=cache,
            )
        )
        del data_obj
        return new_obj

    else:  # make temporary file from lzw compressed COG and load with cache on or off depending on strategy
        dst_profile = cog_profiles.get("lzw")
        delete_file = not cache

        with NamedTemporaryFile(delete=True, suffix=".tif") as in_file:
            with NamedTemporaryFile(delete=delete_file, suffix=".tif") as out_file:
                data_obj.rio.to_raster(in_file.name, tiled=True, windowed=True)
                del data_obj
                cog_translate(in_file.name, out_file.name, dst_profile, in_memory=True)
                return _parse_string_attributes(
                    rxr.open_rasterio(
                        out_file.name,
                        mask_and_scale=True,
                        band_as_variable=band_as_var,
                        cache=cache,
                    )
                )


def _check_dask_array(original_map: Union[xr.DataArray, xr.Dataset]) -> bool:
    """
    Check whether map to be reprojected has dask data or not

    Parameters
    ----------
    original_map : Union[xr.DataArray, xr.Dataset]
        Map to be reprojected

    Returns
    -------
    bool
        Whether the data is a dask array
    """

    chunks = (
        original_map["band_1"].chunks
        if isinstance(original_map, xr.Dataset)
        else original_map.chunks
    )
    return chunks is not None


def _parse_string_attributes(
    obj: Union[xr.DataArray, xr.Dataset]
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Parses string attributes stored in rasters

    Parameters
    ----------
    obj : Union[xr.DataArray, xr.Dataset]
        Xarray object with possible string attributes

    Returns
    -------
    Union[xr.DataArray, xr.Dataset]
        Object returned with parsed attributes
    """

    if "pairing_dictionary" in obj.attrs and isinstance(
        obj.attrs["pairing_dictionary"], str
    ):
        eval_str = ast.literal_eval(
            obj.attrs["pairing_dictionary"]
            .replace("nan", '"nan"')
            .replace("np.int64(", "")
            .replace("np.float64(", "")
            .replace("np.int32(", "")
            .replace("np.float32(", "")
            .replace("np.int16(", "")
            .replace("np.int8(", "")
            .replace("),", ",")
            .replace("))", ")")
            .replace(")}", "}")
        )
        obj.attrs["pairing_dictionary"] = {
            (float(k[0]), float(k[1])): float(v) for k, v in eval_str.items()
        }

        if isinstance(obj, xr.Dataset):
            for var in obj.data_vars:
                obj[var].attrs["pairing_dictionary"] = obj.attrs["pairing_dictionary"]

    if isinstance(obj, xr.Dataset):
        for var in obj.data_vars:
            if "pairing_dictionary" in obj[var].attrs and isinstance(
                obj[var].attrs["pairing_dictionary"], str
            ):
                eval_str = ast.literal_eval(
                    obj[var].attrs["pairing_dictionary"].replace("nan", '"nan"')
                )
                obj[var].attrs["pairing_dictionary"] = {
                    (float(k[0]), float(k[1])): float(v) for k, v in eval_str.items()
                }

    return obj


def _convert_to_dataset(xr_object=Union[xr.DataArray, xr.Dataset]) -> xr.Dataset:
    """
    Converts xarray object to dataset if it is not one already.

    Parameters
    ----------
    xr_object : Union[xr.DataArray, xr.Dataset]
        Xarray object to convert or simply return

    Returns
    -------
    xr.Dataset
        Dataset object

    """

    if isinstance(xr_object, xr.DataArray):
        nodata = xr_object.rio.nodata

        xr_object = (
            xr_object.to_dataset(dim="band")
            if "band" in xr_object.dims
            else xr_object.to_dataset(name="1")
        )
        xr_object = xr_object.rename_vars({x: f"band_{x}" for x in xr_object.data_vars})

        # Account for nodata
        for var_name in xr_object.data_vars:
            xr_object[var_name] = xr_object[var_name].rio.write_nodata(nodata)

        return xr_object
    else:
        return xr_object


def stac_to_df(
    stac_items: ItemCollection,
    assets: list = None,
    attribute_allow_list: list = None,
    attribute_block_list: list = None,
) -> pd.DataFrame:
    """Convert STAC Items in to a DataFrame

    Parameters
    ----------
    stac_items: ItemCollection
        STAC Item Collection returned from pystac client
    assets : list, default = None
        Assets to keep, (keep all if None)
    attribute_allow_list: list, default = None
        List of columns to allow in the result DataFrame
    attribute_block_list: list, default = None
        List of columns to remove in the result DataFrame

    Returns
    -------
    pd.DataFrame
        A DataFrame with rows for each unique item/asset combination

    Raises
    ------
    ValueError
        Allow and block lists should be mutually exclusive
    ValueError
        No entries in DataFrame due to nonexistent asset
    ValueError
        There are no assets in this query to run a catalog comparison

    """

    item_dfs, compare_idx = [], 1

    # Check for mutually exclusive lists
    if (
        len(
            list(
                (
                    Counter(attribute_allow_list) & Counter(attribute_block_list)
                ).elements()
            )
        )
        > 0
    ):
        raise ValueError(
            "There are no assets in this query to run a catalog comparison"
        )

    # Iterate through each STAC Item and make a unique row for each asset
    for item in stac_items:
        item_dict = item.to_dict()
        item_df = pd.json_normalize(item_dict)
        mask = item_df.columns.str.contains("assets.*")
        og_df = item_df.loc[:, ~mask]

        if (
            assets is not None
            and np.sum([asset not in item_dict["assets"].keys() for asset in assets])
            > 0
        ):
            raise ValueError("Non existent asset in parameter assets")

        dfs = []

        # Make a unique row for each asset
        for key, val in item_dict["assets"].items():
            if assets is None or key in assets:
                df = pd.json_normalize(val)
                df["asset"] = key
                df["compare_id"] = compare_idx
                df["map_id"] = val["href"]
                compare_idx += 1
                concat_df = pd.concat([og_df, df], axis=1)
                dfs.append(concat_df.loc[:, ~concat_df.columns.duplicated()])

        if len(dfs) < 1:
            raise ValueError(
                "There are no assets in this query to run a catalog comparison.  "
                "Please revisit original query."
            )

        item_dfs.append(pd.concat(dfs, ignore_index=True))

    # Concatenate the DataFrames and remove unwanted columns if allow and block lists exist
    catalog_df = pd.concat(item_dfs, ignore_index=True)

    if attribute_allow_list is not None:
        catalog_df = catalog_df[attribute_allow_list]

    if attribute_block_list is not None:
        catalog_df = catalog_df.drop(attribute_block_list, axis=1)

    return catalog_df


def _create_circle_mask(
    sizes: int | Tuple[int], center: Tuple[Number, Number], radius: Number
) -> np.ndarray:
    """
    Function to create a circle mask

    Parameters
    ----------
    sizes : Int or Tuple of Int
        Size of xarray's in number of pixels or tuple with the x and y sizes, respectively.
    center : Tuple of Number
        Tuple with the center coordinates (x, y).
    radius : Number
        Radius of the circle.

    Returns
    -------
    mask : np.ndarray
        Numpy array with the circle mask.
    """
    Y, X = np.ogrid[: sizes[1], : sizes[0]]
    dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)
    mask = dist_from_center <= radius
    return mask


def _create_xarray(
    upper_left: Tuple[Number, Number],
    lower_right: Tuple[Number, Number],
    sizes: int | Tuple[int, int],
    band_params: Iterable[Tuple[Number, Number, Tuple[Number, Number], Number]],
    nodata_value: Optional[Number] = None,
    encoded_nodata_value: Optional[Number] = None,
    shapes: str = "circle",
    band_dim_name: str = "band",
    return_dataset: bool = False,
) -> xr.DataArray | xr.Dataset:
    """
    Function to create xarray's with circles per band, and optionally return a Dataset

    Parameters
    ----------
    upper_left : Tuple of Number
        Tuple with the upper left coordinates (x, y).
    lower_right : Tuple of Number
        Tuple with the lower right coordinates (x, y).
    sizes : Int or Tuple of Int
        Size of xarray's in number of pixels or tuple with the x and y sizes, respectively.
    band_params : Iterable of Tuples
        Each tuple represents the parameters of each band comprised of:
        (background_value, circle_value, circle_center, circle_radius).

        Circle center is a tuple with (x, y) coordinates.

        The data types are:
            background_value: Number
            circle_value: Number
            circle_center: tuple of Number
            circle_radius: Number
    nodata_value : Optional[Number], default = None
        Nodata value to use for the xarray's.
    encoded_nodata_value : Optional[Number], default = None
        Nodata value to use for the encoded xarray's.
    shapes : str, default = 'circle'
        Shape of mask within the xarray's. Currently only 'circle' is supported.
    band_dim_name : str, default = 'band'
        Name of the band dimension.
    return_dataset : bool, default = False
        Whether to return a Dataset instead of a DataArray.

    Raises
    ------
    ValueError
        If the shape is not supported.

    Returns
    -------
    data_array : xarray.DataArray or xarray.Dataset
        Xarray with the circles per band.
    """

    # Handle sizes
    if isinstance(sizes, Number):
        sizes = int(sizes)
        sizes = (sizes, sizes)

    # handle shapes
    if shapes == "circle":
        mask_func = _create_circle_mask
    # elif False:
    # placeholder for future shapes
    # pass
    else:
        raise ValueError(f"Shape '{shapes}' is not supported.")

    # Create empty array
    array = np.full((len(band_params),) + tuple(reversed(sizes)), nodata_value)

    for band_idx, params in enumerate(band_params):
        background_value, circle_value, circle_center, circle_radius = params

        # Apply background value, leaving the border as nodata
        array[band_idx, 1:-1, 1:-1] = background_value

        if shapes == "circle":
            # Create circle mask and apply circle value
            circle_mask = mask_func(sizes, circle_center, circle_radius)

            array[band_idx, circle_mask] = circle_value

        # elif False:
        # placeholder for future shapes
        #    pass

    # create xarray DataArray
    data_array = xr.DataArray(data=array, dims=[band_dim_name, "y", "x"])
    data_array.rio.write_crs("epsg:4326", inplace=True)

    # assign nodata values
    if nodata_value is not None:
        data_array.rio.write_nodata(nodata_value, inplace=True)

    if encoded_nodata_value is not None:
        data_array.rio.write_nodata(encoded_nodata_value, inplace=True, encoded=True)

    # create coordinates
    lats = np.linspace(upper_left[1], lower_right[1], sizes[1])
    longs = np.linspace(lower_right[0], upper_left[0], sizes[0])
    bands = np.arange(1, len(band_params) + 1)

    # assign coordinates
    data_array = data_array.assign_coords({band_dim_name: bands, "y": lats, "x": longs})

    if return_dataset:
        # Convert the DataArray to a Dataset
        dataset = data_array.to_dataset(name="variable")
        return dataset

    return data_array


def _create_xarray_pairs(
    upper_left: Tuple[Number, Number],
    lower_right: Tuple[Number, Number],
    sizes: int | Tuple[int, int],
    band_params_candidate: Iterable[
        Tuple[Number, Number, Tuple[Number, Number], Number]
    ],
    band_params_benchmark: Iterable[
        Tuple[Number, Number, Tuple[Number, Number], Number]
    ],
    nodata_value: Optional[Number] = None,
    encoded_nodata_value: Optional[Number] = None,
    shapes: str = "circle",
    band_dim_name: str = "band",
    return_dataset: bool = False,
) -> Tuple[xr.DataArray | xr.Dataset]:
    """
    Function to create xarray's with shapes per band, and optionally return a Dataset

    Parameters
    ----------
    upper_left : Tuple of Number
        Tuple with the upper left coordinates (x, y).
    lower_right : Tuple of Number
        Tuple with the lower right coordinates (x, y).
    sizes : Int or Tuple of Int
        Size of xarray's in number of pixels or tuple with the x and y sizes, respectively.
    band_params : Iterable of Tuples
        Each tuple represents the parameters of each band comprised of:
        (background_value, circle_value, circle_center, circle_radius).

        Circle center is a tuple with (x, y) coordinates.

        The data types are:
            background_value: Number
            circle_value: Number
            circle_center: tuple of Number
            circle_radius: Number
    nodata_value : Optional[Number], default = None
        Nodata value to use for the xarray's.
    encoded_nodata_value : Optional[Number], default = None
        Nodata value to use for the encoded xarray's.
    shapes : str, default = 'circle'
        Shape of mask within the xarray's. Currently only 'circle' is supported.
    band_dim_name : str, default = 'band'
        Name of the band dimension.
    return_dataset : bool, default = False
        Whether to return a Dataset instead of a DataArray.

    Raises
    ------
    ValueError
        If the shape is not supported.

    Returns
    -------
    data_array : xarray.DataArray or xarray.Dataset
        Candidate map.
    data_array : xarray.DataArray or xarray.Dataset
        Benchmark map.
    """

    args = [
        upper_left,
        lower_right,
        sizes,
        None,
        nodata_value,
        encoded_nodata_value,
        shapes,
        band_dim_name,
        return_dataset,
    ]

    candidate_args = args.copy()
    benchmark_args = args.copy()

    candidate_args[3] = band_params_candidate
    benchmark_args[3] = band_params_benchmark

    candidate_map = _create_xarray(*candidate_args)
    benchmark_map = _create_xarray(*benchmark_args)

    return candidate_map, benchmark_map



"""
Compare catalogs of candidates and benchmarks.
"""
from __future__ import annotations

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

import gc
from typing import Iterable, Optional, Callable, Tuple
import os

import pandas as pd
from rioxarray import open_rasterio as rxr_or
import xarray as xr
import dask.dataframe as dd


def catalog_compare(
    candidate_catalog: pd.DataFrame | dd.DataFrame,
    benchmark_catalog: pd.DataFrame | dd.DataFrame,
    map_ids: str | Iterable[str],
    how: str = "inner",
    on: Optional[str | Iterable[str]] = None,
    left_on: Optional[str | Iterable[str]] = None,
    right_on: Optional[str | Iterable[str]] = None,
    suffixes: tuple[str, str] = ("_candidate", "_benchmark"),
    merge_kwargs: Optional[dict] = None,
    open_kwargs: Optional[dict] = None,
    compare_type: str | Callable = "continuous",
    compare_kwargs: Optional[dict] = None,
    agreement_map_field: Optional[str] = None,
    agreement_map_write_kwargs: Optional[dict] = None,
) -> pd.DataFrame | dd.DataFrame:
    """
    Compare catalogs of candidate and benchmark maps.

    Parameters
    ----------
    candidate_catalog : pandas.DataFrame | dask.DataFrame
        Candidate catalog.
    benchmark_catalog : pandas.DataFrame | dask.DataFrame
        Benchmark catalog.
    map_ids : str | Iterable of str
        Column name(s) where maps or paths to maps occur. If str is given, then the same value should occur in both catalogs. If Iterable[str] is given of length 2, then the column names where maps are will be in [candidate, benchmark] respectively.

        The columns corresponding to map_ids should have either str, xarray.DataArray, xarray.Dataset, rasterio.io.DatasetReader, rasterio.vrt.WarpedVRT, or os.PathLike objects.
    how : str, default = "inner"
        Type of merge to perform. See pandas.DataFrame.merge for more information.
    on : str | Iterable of str, default = None
        Column(s) to join on. Must be found in both catalogs. If None, and left_on and right_on are also None, then the intersection of the columns in both catalogs will be used.
    left_on : str | Iterable of str, default = None
        Column(s) to join on in left catalog. Must be found in left catalog.
    right_on : str | Iterable of str, default = None
        Column(s) to join on in right catalog. Must be found in right catalog.
    suffixes : tuple of str, default = ("_candidate", "_benchmark")
        Suffixes to apply to overlapping column names in candidate and benchmark catalogs, respectively. Length two tuple of strings.
    merge_kwargs : dict, default = None
        Keyword arguments to pass to pandas.DataFrame.merge.
    compare_type : str | Callable, default = "continuous"
        Type of comparison to perform. If str, then must be one of {"continuous", "categorical", "probabilistic"}. If Callable, then must be a function that takes two xarray.DataArray or xarray.Dataset objects and returns a tuple of length 2. The first element of the tuple must be an xarray.DataArray or xarray.Dataset object representing the agreement map. The second element of the tuple must be a pandas.DataFrame object representing the metrics.
    compare_kwargs : dict, default = None
        Keyword arguments to pass to the compare_type function.
    agreement_map_field : str, default = None
        Column name to write agreement maps to. If None, then agreement maps will not be written to file.
    agreement_map_write_kwargs : dict, default = None
        Keyword arguments to pass to xarray.DataArray.rio.to_raster when writing agreement maps to file.

    Raises
    ------
    ValueError
        If map_ids is not str or Iterable of str.
        If compare_type is not str or Callable.
        If compare_type is str and not one of {"continuous", "categorical", "probabilistic"}.
    NotImplementedError
        If compare_type is "probabilistic".

    Returns
    -------
    pandas.DataFrame | dask.DataFrame
        Agreement catalog.
    """

    # unpack map_ids
    if isinstance(map_ids, str):
        candidate_map_ids, benchmark_map_ids = map_ids, map_ids
    elif isinstance(map_ids, Iterable):
        candidate_map_ids, benchmark_map_ids = map_ids
    else:
        raise ValueError("map_ids must be str or Iterable of str")

    # set merge_kwargs to empty dict if None
    if merge_kwargs is None:
        merge_kwargs = dict()

    # create agreement catalog
    agreement_catalog = candidate_catalog.merge(
        benchmark_catalog,
        how=how,
        on=on,
        left_on=left_on,
        right_on=right_on,
        suffixes=suffixes,
        **merge_kwargs,
    )

    def compare_row(
        row,
        compare_type: str | Callable,
        compare_kwargs: dict,
        open_kwargs: dict,
        agreement_map_field: str,
        agreement_map_write_kwargs: dict,
    ) -> Tuple[xr.DataArray | xr.Dataset, pd.DataFrame]:
        """Compares catalog and benchmark maps by rows"""

        def loadxr(map, open_kwargs):
            """load xarray object if not already"""
            return (
                map
                if isinstance(map, (xr.DataArray, xr.Dataset))
                else rxr_or(map, **open_kwargs)
            )

        # load maps
        candidate_map = loadxr(row[candidate_map_ids + suffixes[0]], open_kwargs)
        benchmark_map = loadxr(row[benchmark_map_ids + suffixes[1]], open_kwargs)

        # set compare_kwargs to empty dict if None
        if compare_kwargs is None:
            compare_kwargs = dict()

        # set agreement_map_write_kwargs to empty dict if None
        if agreement_map_write_kwargs is None:
            agreement_map_write_kwargs = dict()

        if isinstance(compare_type, str):
            if compare_type == "categorical":
                results = candidate_map.gval.categorical_compare(
                    benchmark_map, **compare_kwargs
                )

                # results is a tuple of length 3 or 4
                # agreement_map, crosstab_df, metrics_df, attrs_df = results
                # where attrs_df is optional
                agreement_map, metrics_df = results[0], results[2]

            elif compare_type == "continuous":
                results = candidate_map.gval.continuous_compare(
                    benchmark_map, **compare_kwargs
                )

                # results is a tuple of length 2 or 3
                # agreement_map, metrics_df, attrs_df = results
                # where attrs_df is optional
                agreement_map, metrics_df = results[:2]

            elif compare_type == "probabilistic":
                raise NotImplementedError(
                    "probabilistic comparison not implemented yet"
                )

            else:
                raise ValueError(
                    "compare_type of type str must be one of {'continuous', 'categorical', 'probabilistic'}"
                )

        elif isinstance(compare_type, Callable):
            agreement_map, metrics_df = compare_type(
                candidate_map, benchmark_map, **compare_kwargs
            )

        else:
            raise ValueError("compare_type must be str or Callable")

        # Write agreement map to file
        if (agreement_map_field is not None) & isinstance(
            agreement_map, (xr.DataArray, xr.Dataset)
        ):
            if isinstance(row[agreement_map_field], (str, os.PathLike)):
                agreement_map.rio.to_raster(
                    row[agreement_map_field], **agreement_map_write_kwargs
                )

        # Unfortunately necessary until a fix is found in xarray/rioxarray io
        del candidate_map, benchmark_map, agreement_map
        gc.collect()

        return metrics_df

    # make kwargs for dask apply
    if isinstance(agreement_catalog, dd.DataFrame):
        dask_kwargs = {"meta": ("output", "f8")}
    else:
        dask_kwargs = {}

    # set open_kwargs to empty dict if None
    if open_kwargs is None:
        open_kwargs = dict()

    # apply compare_row to each row of agreement_catalog
    metrics_df = agreement_catalog.apply(
        compare_row,
        axis=1,
        **dask_kwargs,
        compare_type=compare_type,
        open_kwargs=open_kwargs,
        compare_kwargs=compare_kwargs,
        agreement_map_field=agreement_map_field,
        agreement_map_write_kwargs=agreement_map_write_kwargs,
    )

    def nested_merge(i, sub_df) -> pd.DataFrame:
        """Duplicated agreement row for each row in sub_df"""
        try:
            agreement_row = agreement_catalog.iloc[i].to_frame().T
        except NotImplementedError:
            agreement_row = agreement_catalog.loc[agreement_catalog.index == i]

        sub_df.index = [i] * len(sub_df)
        return agreement_row.join(sub_df)

    # merge agreement_catalog with metrics_df
    if isinstance(metrics_df, dd.Series):
        return dd.concat(
            [nested_merge(i, sub_df) for i, sub_df in enumerate(metrics_df)]
        ).reset_index(drop=True)

    if isinstance(metrics_df, pd.Series):
        return pd.concat(
            [nested_merge(i, sub_df) for i, sub_df in enumerate(metrics_df)]
        ).reset_index(drop=True)



"""
Test functionality for gval/compare.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

import numpy as np
import pandas as pd
import pytest
from pytest_cases import parametrize_with_cases
import xarray as xr
from pytest import raises

from gval.comparison.pairing_functions import (
    _is_not_natural_number,
    cantor_pair,
    cantor_pair_signed,
    szudzik_pair,
    szudzik_pair_signed,
    _make_pairing_dict,
    _make_pairing_dict_fn,
    PairingDict,
    difference,
)
from gval.comparison.agreement import _compute_agreement_map
from gval.comparison.tabulation import (
    _crosstab_2d_DataArrays,
    _crosstab_3d_DataArrays,
    _crosstab_DataArrays,
    _crosstab_Datasets,
)
from gval.comparison.compute_comparison import ComparisonProcessing
from gval.utils.schemas import Crosstab_df

from tests.conftest import _assert_pairing_dict_equal


compare_proc = ComparisonProcessing()


@parametrize_with_cases("number, expected", glob="is_not_natural_number_successes")
def test_is_not_natural_number_success(number, expected):
    assert expected == _is_not_natural_number(number, False), "Expected natural number"


@parametrize_with_cases("number", glob="is_not_natural_number_failures")
def test_is_not_natural_number_failures(number):
    with pytest.raises(ValueError):
        _is_not_natural_number(number, True)


@parametrize_with_cases("c, b, a", glob="cantor_pair")
def test_cantor_pair(c, b, a):
    """tests cantor pairing function"""
    np.testing.assert_equal(
        cantor_pair(c, b), a
    ), "Cantor function output does not match expected value"


@parametrize_with_cases("c, b, a", glob="cantor_pair_signed")
def test_cantor_pair_signed(c, b, a):
    """
    tests cantor pairing function
    FIXME: Figure out why RuntimeWarning is occurring for test cases with np.nan.
    """
    np.testing.assert_equal(
        cantor_pair_signed(c, b), a
    ), "Signed cantor function output does not match expected value"


@parametrize_with_cases("c, b, a", glob="szudzik_pair")
def test_szudzik_pair(c, b, a):
    np.testing.assert_equal(
        szudzik_pair(c, b), a
    ), "szudzik function output does not match expected value"


@parametrize_with_cases("c, b, a", glob="szudzik_pair_signed")
def test_szudzik_pair_signed(c, b, a):
    """
    Tests szudzik pairing function
    FIXME: Figure out why RuntimeWarning is occurring for test cases with np.nan.
    """
    np.testing.assert_equal(
        szudzik_pair_signed(c, b), a
    ), "Signed szudzik function output does not match expected value"


@parametrize_with_cases("dict_with_nans, expected", glob="replace_nans_in_pairing_dict")
def test_replace_nans_in_pairing_dict(dict_with_nans, expected):
    _assert_pairing_dict_equal(PairingDict(dict_with_nans), expected)


@parametrize_with_cases(
    "unique_candidate_values, unique_benchmark_values, expected_dict",
    glob="make_pairing_dict",
)
def test_make_pairing_dict(
    unique_candidate_values, unique_benchmark_values, expected_dict
):
    """Tests creation of a pairing dictionary"""

    computed_dict = _make_pairing_dict(unique_candidate_values, unique_benchmark_values)
    expected_dict = PairingDict(expected_dict)

    _assert_pairing_dict_equal(computed_dict, expected_dict)


@parametrize_with_cases("c, b, pairing_dict, expected_value", glob="pairing_dict_fn")
def test_pairing_dict_fn(c, b, pairing_dict, expected_value):
    """Tests pairing dictionary function"""

    pairing_dict_fn = _make_pairing_dict_fn(pairing_dict)
    computed_value = pairing_dict_fn(c, b)

    np.testing.assert_equal(computed_value, expected_value)


@parametrize_with_cases("c, b, expected_value", glob="difference")
def test_difference(c, b, expected_value):
    """Tests difference comparison function"""

    np.testing.assert_equal(difference(c, b), expected_value)


#################################################################################


@parametrize_with_cases("agreement_map, expected_df", glob="crosstab_2d_DataArrays")
def test_crosstab_2d_DataArrays(agreement_map, expected_df):
    """Test crosstabbing agreement DataArrays"""
    agreement_map.gval  # Initialize gval for attributes
    crosstab_df = _crosstab_2d_DataArrays(agreement_map)
    pd.testing.assert_frame_equal(crosstab_df, expected_df, check_dtype=False)


@parametrize_with_cases("agreement_map, expected_df", glob="crosstab_3d_DataArrays")
def test_crosstab_3d_DataArrays(agreement_map, expected_df):
    """Test crosstabbing agreement DataArrays"""
    agreement_map.gval  # Initialize gval for attributes
    crosstab_df = _crosstab_3d_DataArrays(agreement_map)
    pd.testing.assert_frame_equal(crosstab_df, expected_df, check_dtype=False)


@parametrize_with_cases(
    "agreement_map, expected_df", glob="crosstab_DataArrays_success"
)
def test_crosstab_DataArrays_success(agreement_map, expected_df):
    """Test crosstabbing agreement DataArrays"""
    agreement_map.gval  # Initialize gval for attributes
    crosstab_df = _crosstab_DataArrays(agreement_map)
    pd.testing.assert_frame_equal(crosstab_df, expected_df, check_dtype=False)


@parametrize_with_cases("agreement_map", glob="crosstab_DataArrays_fail")
def test_crosstab_DataArrays_fail(agreement_map):
    """Test crosstabbing agreement DataArrays"""
    agreement_map.gval  # Initialize gval for attributes
    with raises(ValueError):
        _crosstab_DataArrays(agreement_map)


@parametrize_with_cases("agreement_map, expected_df", glob="crosstab_Datasets")
def test_crosstab_Datasets(agreement_map, expected_df):
    """Test crosstabbing agreement datasets"""
    agreement_map.gval  # Initialize gval for attributes
    crosstab_df = _crosstab_Datasets(agreement_map)
    # takes band_# pattern to just #
    crosstab_df["band"] = crosstab_df["band"].apply(lambda x: x.split("_")[-1])
    crosstab_df = Crosstab_df(crosstab_df)
    pd.testing.assert_frame_equal(crosstab_df, expected_df, check_dtype=False)


@parametrize_with_cases(
    "candidate_map, benchmark_map, agreement_map, comparison_function, pairing_dict, allow_candidate_values, allow_benchmark_values, nodata, encode_nodata",
    glob="compute_agreement_map_success",
)
def test_compute_agreement_map_success(
    candidate_map,
    benchmark_map,
    agreement_map,
    comparison_function,
    pairing_dict,
    allow_candidate_values,
    allow_benchmark_values,
    nodata,
    encode_nodata,
):
    """Tests computing of agreement xarray from two xarrays"""

    agreement_map_computed = _compute_agreement_map(
        candidate_map,
        benchmark_map,
        comparison_function,
        pairing_dict,
        allow_candidate_values=allow_candidate_values,
        allow_benchmark_values=allow_benchmark_values,
        nodata=nodata,
        encode_nodata=encode_nodata,
    )

    # Use xr.testing.assert_identical() if names and attributes need to be compared too
    xr.testing.assert_equal(agreement_map_computed, agreement_map)


@parametrize_with_cases(
    "candidate_map, benchmark_map, comparison_function, allow_candidate_values, allow_benchmark_values, nodata, encode_nodata",
    glob="compute_agreement_map_fail",
)
def test_compute_agreement_map_fail(
    candidate_map,
    benchmark_map,
    comparison_function,
    allow_candidate_values,
    allow_benchmark_values,
    nodata,
    encode_nodata,
):
    """Tests fail computing of agreement xarray from two xarrays"""

    with raises(ValueError):
        _ = _compute_agreement_map(
            candidate_map,
            benchmark_map,
            comparison_function,
            allow_candidate_values=allow_candidate_values,
            allow_benchmark_values=allow_benchmark_values,
            nodata=nodata,
            encode_nodata=encode_nodata,
        )


@parametrize_with_cases(
    "candidate_map, benchmark_map, agreement_map, comparison_function, allow_candidate_values, allow_benchmark_values, nodata, encode_nodata",
    glob="comparison_processing_agreement_maps_success",
)
def test_comparison_processing_agreement_map_success(
    candidate_map,
    benchmark_map,
    agreement_map,
    comparison_function,
    allow_candidate_values,
    allow_benchmark_values,
    nodata,
    encode_nodata,
):
    """Tests comparison processing computing of agreement xarray from two xarrays"""

    agreement_map_computed = compare_proc.process_agreement_map(
        comparison_function=comparison_function,
        candidate_map=candidate_map,
        benchmark_map=benchmark_map,
        allow_candidate_values=allow_candidate_values,
        allow_benchmark_values=allow_benchmark_values,
        nodata=nodata,
        encode_nodata=encode_nodata,
    )

    # Use xr.testing.assert_identical() if names and attributes need to be compared too
    xr.testing.assert_equal(agreement_map_computed, agreement_map)


@parametrize_with_cases(
    "candidate_map, benchmark_map, comparison_function, allow_candidate_values, allow_benchmark_values, nodata, encode_nodata, exception",
    glob="comparison_processing_agreement_maps_fail",
)
def test_comparison_processing_agreement_maps_fail(
    candidate_map,
    benchmark_map,
    comparison_function,
    allow_candidate_values,
    allow_benchmark_values,
    nodata,
    encode_nodata,
    exception,
):
    with raises(exception):
        _ = compare_proc.process_agreement_map(
            comparison_function=comparison_function,
            candidate_map=candidate_map,
            benchmark_map=benchmark_map,
            allow_candidate_values=allow_candidate_values,
            allow_benchmark_values=allow_benchmark_values,
            nodata=nodata,
            encode_nodata=encode_nodata,
        )


def test_comparison_get_all_param():
    """tests get all params function"""

    try:
        compare_proc.get_all_parameters()
    except KeyError:
        assert False, "Signature dict not present or keys changed"


def test_get_available_functions():
    """tests get all available functions"""

    assert compare_proc.available_functions() == [
        "pairing_dict",
        "cantor",
        "szudzik",
        "difference",
    ]


@parametrize_with_cases("args, func", glob="comparison_register_function")
def test_comparison_register_function(args, func):
    """tests register func function"""

    compare_proc.register_function(**args)(func)


@parametrize_with_cases(
    "args, func, exception", glob="comparison_register_function_fail"
)
def test_comparison_register_function_fail(args, func, exception):
    """tests register func fail function"""

    with raises(exception):
        compare_proc.register_function(**args)(func)


@parametrize_with_cases("names, args, cls", glob="comparison_register_class")
def test_register_class(names, args, cls):
    """tests register class function"""

    compare_proc.register_function_class(**args)(cls)

    if [name in compare_proc.registered_functions for name in names] != [True] * len(
        names
    ):
        assert False, "Unable to register all class functions"


@parametrize_with_cases("args, cls, exception", glob="comparison_register_class_fail")
def test_comparison_register_class_fail(args, cls, exception):
    """tests register class fail function"""

    with raises(exception):
        compare_proc.register_function_class(**args)(cls)


@parametrize_with_cases("name, params", glob="comparison_get_param")
def test_comparison_get_param(name, params):
    """tests get param function"""

    _params = compare_proc.get_parameters(name)
    assert _params == params


@parametrize_with_cases("name", glob="comparison_get_param_fail")
def test_comparison_get_param_fail(name):
    """tests get param fail function"""

    with raises(KeyError):
        _ = compare_proc.get_parameters(name)



import warnings
from typing import Tuple, Union
from itertools import zip_longest

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
import xarray as xr
import contextily as cx
import xyzservices


def _map_plot(
    ds: Union[xr.DataArray, xr.Dataset],
    title: str = "Categorical Map",
    colormap: str = "viridis",
    figsize: Tuple[int, int] = None,
    legend_labels: list = None,
    plot_bands: Union[str, list] = "all",
    plot_type: str = "categorical",
    colorbar_label: Union[str, list] = "",
    basemap: xyzservices.lib.TileProvider = cx.providers.OpenStreetMap.Mapnik,
):
    """
    Plots categorical or continuous Map for xarray object

    Parameters
    __________
    ds : xr.DataArray
        Dataset to create categorical map
    title : str
        Title of map, default = "Categorical Map"
    colormap : str, default = "viridis"
        Colormap of data
    figsize : tuple[int, int], default=None
        Size of the plot
    legend_labels : list, default = None
        Override labels in legend
    plot_bands : Union[str, list], default = 'all'
        Which bands to plot if multiple. Default is all bands.
    plot_type : str, default = 'categorical', options
        Whether to plot the map as a categorical map
    color_bar_label : Union[str, list], default =""
        Label or labels for colorbar in the case of continuous plots
    basemap : Union[bool, xyzservices.lib.TileProvider], default = cx.providers.Stamen.Terrain
        Add basemap to the plot

    Returns
    -------
    Union[xr.DataArray, xr.Dataset]
        QuadMesh Matplotlib object

    Raises
    ------
    ValueError
        Too many values present in dataset for categorical plot.

    References
    ----------
    .. [1] `Matplotlib figure <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html>`_
    .. [2] `Matplotlib legend <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html>`_
    """

    categorical = True if plot_type == "categorical" else False

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        # Get datasets
        if isinstance(ds, xr.Dataset):
            ds_list = [
                ds[x]
                for x in ds.data_vars
                if plot_bands == "all" or int(x.split("_")[1]) in plot_bands
            ]
            bands = (
                [x + 1 for x in range(len(ds.data_vars))]
                if plot_bands == "all"
                else plot_bands
            )

        else:
            if len(ds.shape) == 3:
                ds_list = [
                    ds.sel({"band": x + 1})
                    for x in range(ds.shape[0])
                    if plot_bands == "all" or x in plot_bands
                ]
                bands = (
                    [x + 1 for x in range(ds.shape[0])]
                    if plot_bands == "all"
                    else plot_bands
                )
            elif len(ds.shape) > 3 or len(ds.shape) < 2:
                raise ValueError("Needs to be 2 or 3 dimensional xarray object")
            else:
                ds_list = [ds]
                bands = [1]

        if len(ds_list) > 8:
            raise ValueError("Cannot plot more than 8 DataArrays at a time")

        cols = 2 if len(ds_list) > 1 else 1
        rows = ((len(ds_list) - 1) // 2) + 1

        if figsize is None:
            figsize = (5 * cols, 4 * rows)

        # Setup figure, axis, and plot
        fig, axs = plt.subplots(rows, cols, figsize=figsize)
        axes = axs.ravel() if "ravel" in dir(axs) else [axs]

        for i, ax in enumerate(axes):
            if i >= len(ds_list):
                continue

            ds_c = ds_list[i]
            plot = ds_c.plot(ax=ax, cmap=colormap)

            # Print CRS if not large WKT String
            crs = ds_c.rio.crs if len(ds_c.rio.crs) < 25 else ""
            ax.set_title(f"Band {bands[i]}")
            ax.set_xlabel("Longitude")
            ax.set_ylabel("Latitude")

            if basemap:  # pragma: no cover
                cx.add_basemap(ax, crs=ds_c.rio.crs, source=basemap)

            if categorical:
                # Get colormap values for each unique value
                cmap = matplotlib.colormaps[colormap]
                unique_vals = np.unique(ds_c)

                if len(unique_vals) > 25:
                    raise ValueError(
                        "Too many values present in dataset for categorical plot."
                    )

                if legend_labels is not None and len(legend_labels) != len(unique_vals):
                    raise ValueError("Need as many labels as unique values.")

                # Setup normalized color ramp
                norm = matplotlib.colors.Normalize(
                    vmin=np.nanmin(unique_vals), vmax=np.nanmax(unique_vals)
                )

                # Create legend
                labels = unique_vals if legend_labels is None else legend_labels
                legend_elements = [
                    Patch(color=cmap(norm(val)), label=str(label))
                    for val, label in zip(unique_vals, labels)
                    if not np.isnan(val)
                ]
                ax.legend(
                    title="Encodings",
                    handles=legend_elements,
                    bbox_to_anchor=(1.05, 1.0),
                    loc="upper left",
                )

        fig.suptitle(f"{title} ({crs})", fontsize=15)

        if len(ds_list) <= 2:
            top = 0.85
        elif len(ds_list) <= 4:
            top = 0.9
        elif len(ds_list) <= 6:
            top = 0.925
        else:
            top = 0.95

        plt.subplots_adjust(
            top=top, bottom=0.1, left=0.125, right=0.9, hspace=0.6, wspace=0.5
        )

        if categorical:
            # Erase color bar and autoformat x labels to not overlap
            while len(fig.axes) > len(ds_list):
                fig.delaxes(fig.axes[len(ds_list)])
        else:
            # Erase extra axis if present
            if len(ds_list) % 2 == 1 and len(ds_list) > 1:
                fig.delaxes(fig.axes[len(ds_list)])

            use_label = None
            for idx, label in zip_longest(
                range(len(ds_list) // 2, len(ds_list)), colorbar_label
            ):
                # If less labels are provided than graphs apply last label to the remaining graphs
                use_label = use_label if label is None else label
                fig.axes[idx].set_ylabel(label)

        fig.autofmt_xdate()

        fig.show()

        return plot



"""
Carries over attributes from xarray's DataArray and Dataset classes.

TODO:
    - [ ] implement tests with xr.testing.assert_identical
    - [ ] make additional test cases???
    - [ ] add to accessor methods
    - [ ] documentation
"""

__author__ = "Fernando Aristizabal"

from typing import Optional, Iterable, Union, Tuple

import pandas as pd
import xarray as xr
from pandera.typing import DataFrame

from gval.utils.schemas import AttributeTrackingDf


def _attribute_tracking_xarray(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    agreement_map: Optional[Union[xr.DataArray, xr.Dataset]] = None,
    candidate_suffix: Optional[str] = "_candidate",
    benchmark_suffix: Optional[str] = "_benchmark",
    candidate_include: Optional[Iterable[str]] = None,
    candidate_exclude: Optional[Iterable[str]] = None,
    benchmark_include: Optional[Iterable[str]] = None,
    benchmark_exclude: Optional[Iterable[str]] = None,
) -> Union[
    DataFrame[AttributeTrackingDf],
    Tuple[DataFrame[AttributeTrackingDf], Union[xr.DataArray, xr.Dataset]],
]:
    """
    Concatenate xarray attributes into a single pandas dataframe.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map xarray object.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map xarray object.
    candidate_suffix : Optional[str], default = '_candidate'
        Suffix to append to candidate map xarray attributes, by default '_candidate'.
    benchmark_suffix : Optional[str], default = '_benchmark'
        Suffix to append to benchmark map xarray attributes, by default '_benchmark'.
    candidate_include : Optional[Iterable[str]], default = None
        List of attributes to include from candidate map. candidate_include and candidate_exclude are mutually exclusive arguments.
    candidate_exclude : Optional[Iterable[str]], default = None
        List of attributes to exclude from candidate map. candidate_include and candidate_exclude are mutually exclusive arguments.
    benchmark_include : Optional[Iterable[str]], default = None
        List of attributes to include from benchmark map. benchmark_include and benchmark_exclude are mutually exclusive arguments.
    benchmark_exclude : Optional[Iterable[str]], default = None
        List of attributes to exclude from benchmark map. benchmark_include and benchmark_exclude are mutually exclusive arguments.

    Raises
    ------
    ValueError
        If candidate_include and candidate_exclude are both not None.
    ValueError
        If benchmark_include and benchmark_exclude are both not None.

    Returns
    -------
    Union[DataFrame[AttributeTrackingDf], Tuple[DataFrame[AttributeTrackingDf], Union[xr.DataArray, xr.Dataset]]]
        Pandas dataframe with concatenated attributes from candidate and benchmark maps. If agreement_map is not None, returns a tuple with the dataframe and the agreement map.
    """

    candidate_attrs = candidate_map.attrs
    benchmark_attrs = benchmark_map.attrs

    # remove default exclude from both includes
    if (candidate_include is not None) and (candidate_exclude is not None):
        raise ValueError(
            "candidate_include and candidate_exclude are mutually exclusive"
        )

    if (benchmark_include is not None) and (benchmark_exclude is not None):
        raise ValueError(
            "benchmark_include and benchmark_exclude are mutually exclusive"
        )

    # candidate include and exclude
    if candidate_include is not None:
        candidate_attrs = {
            k: v for k, v in candidate_attrs.items() if k in candidate_include
        }
    elif candidate_exclude is not None:
        candidate_attrs = {
            k: v for k, v in candidate_attrs.items() if k not in candidate_exclude
        }

    # benchmark include and exclude
    if benchmark_include is not None:
        benchmark_attrs = {
            k: v for k, v in benchmark_attrs.items() if k in benchmark_include
        }
    elif benchmark_exclude is not None:
        benchmark_attrs = {
            k: v for k, v in benchmark_attrs.items() if k not in benchmark_exclude
        }

    # Convert xarray attributes to pandas dataframes
    candidate_df = pd.DataFrame([candidate_attrs], index=[0])
    benchmark_df = pd.DataFrame([benchmark_attrs], index=[0])

    # Append a suffix to each dataframe's column names to denote its origin
    candidate_df.columns = [f"{col}{candidate_suffix}" for col in candidate_df.columns]
    benchmark_df.columns = [f"{col}{benchmark_suffix}" for col in benchmark_df.columns]

    # Concatenate the dataframes together
    combined_df = pd.concat([candidate_df, benchmark_df], axis=1)

    # validate schema
    AttributeTrackingDf.validate_column_suffixes(
        combined_df, candidate_suffix, benchmark_suffix
    )

    if agreement_map is None:
        return combined_df

    updated_candidate_attrs = candidate_df.to_dict(orient="records")[0]
    updated_benchmark_attrs = benchmark_df.to_dict(orient="records")[0]

    agreement_map = agreement_map.assign_attrs(
        {**updated_candidate_attrs, **updated_benchmark_attrs}
    )

    return combined_df, agreement_map



"""
Defines gval accessor for xarray objects.
"""

from typing import Iterable, Optional, Tuple, Union, Callable, Dict, List, Any
from numbers import Number

import numpy as np
import numba as nb
import pandas as pd
import xarray as xr
from rasterio.enums import Resampling
from pandera.typing import DataFrame
import geopandas as gpd
import contextily as cx
import xyzservices

from gval.homogenize.spatial_alignment import _spatial_alignment
from gval.homogenize.rasterize import _rasterize_data
from gval.homogenize.vectorize import _vectorize_data
from gval.homogenize.numeric_alignment import _align_numeric_data_type
from gval import Comparison
from gval.comparison.tabulation import _crosstab_Datasets, _crosstab_DataArrays
from gval.comparison.compute_categorical_metrics import _compute_categorical_metrics
from gval.comparison.compute_continuous_metrics import _compute_continuous_metrics
from gval.comparison.compute_probabilistic_metrics import _compute_probabilistic_metrics
from gval.attributes.attributes import _attribute_tracking_xarray
from gval.utils.loading_datasets import _parse_string_attributes
from gval.utils.visualize import _map_plot
from gval.comparison.pairing_functions import difference
from gval.subsampling.subsampling import subsample
from gval.utils.schemas import (
    Crosstab_df,
    Metrics_df,
    AttributeTrackingDf,
    Prob_metrics_df,
)


class GVALXarray:
    """
    Class for extending xarray functionality.

    Attributes
    ----------
    _obj : Union[xr.Dataset, xr.DataArray]
       Object to use off the accessor
    data_type : type
       Data type of the _obj
    """

    def __init__(self, xarray_obj):
        self._obj = _parse_string_attributes(xarray_obj)
        self.data_type = type(xarray_obj)
        self.agreement_map_format = "raster"

    def check_same_type(self, benchmark_map: Union[xr.Dataset, xr.DataArray]):
        """
        Makes sure benchmark map is the same data type as the candidate object

        Parameters
        ----------
        benchmark_map: Union[xr.Dataset, xr.DataArray]
            Benchmark Map

        Raises
        ------
        TypeError

        """
        if not isinstance(benchmark_map, self.data_type):
            raise TypeError(f"Benchmark Map needs to be data type of {self.data_type}")

    def __handle_attribute_tracking(
        self,
        candidate_map: Union[xr.Dataset, xr.DataArray],
        benchmark_map: Union[xr.Dataset, xr.DataArray],
        agreement_map: Optional[Union[xr.Dataset, xr.DataArray]] = None,
        attribute_tracking_kwargs: Optional[Dict] = None,
    ):  # pragma: no cover
        """
        Handles attribute tracking for categorical and continuous comparison
        """

        # use user passed attribute_tracking_kwargs to pass arguments to attribute_tracking_xarray()
        if attribute_tracking_kwargs is not None:
            if "benchmark_map" in attribute_tracking_kwargs:
                del attribute_tracking_kwargs["benchmark_map"]

            if "agreement_map" in attribute_tracking_kwargs:
                if attribute_tracking_kwargs["agreement_map"] is None:
                    agreement_map = None
                else:
                    del attribute_tracking_kwargs["agreement_map"]

            results = _attribute_tracking_xarray(
                candidate_map=candidate_map,
                benchmark_map=benchmark_map,
                agreement_map=agreement_map,
                **attribute_tracking_kwargs,
            )

        else:
            results = _attribute_tracking_xarray(
                candidate_map=candidate_map,
                benchmark_map=benchmark_map,
                agreement_map=agreement_map,
            )

        return results

    @Comparison.comparison_function_from_string
    def categorical_compare(
        self,
        benchmark_map: Union[gpd.GeoDataFrame, xr.Dataset, xr.DataArray],
        positive_categories: Union[
            Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]
        ],
        comparison_function: Union[
            Callable, nb.np.ufunc.dufunc.DUFunc, np.ufunc, np.vectorize, str
        ] = "szudzik",
        metrics: Union[str, Iterable[str]] = "all",
        target_map: Optional[Union[xr.Dataset, str]] = "benchmark",
        resampling: Optional[Resampling] = Resampling.nearest,
        pairing_dict: Optional[Dict[Tuple[Number, Number], Number]] = None,
        allow_candidate_values: Optional[Iterable[Union[int, float]]] = None,
        allow_benchmark_values: Optional[Iterable[Union[int, float]]] = None,
        nodata: Optional[Number] = None,
        encode_nodata: Optional[bool] = False,
        exclude_value: Optional[Number] = None,
        negative_categories: Union[
            Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]
        ] = None,
        average: str = "micro",
        weights: Optional[Iterable[Number]] = None,
        rasterize_attributes: Optional[list] = None,
        attribute_tracking: bool = False,
        attribute_tracking_kwargs: Optional[Dict] = None,
        subsampling_df: Optional[gpd.GeoDataFrame] = None,
        subsampling_average: Optional[str] = None,
    ) -> Tuple[
        Union[
            Tuple[
                Union[xr.Dataset, xr.DataArray],
                DataFrame[Crosstab_df],
                DataFrame[Metrics_df],
            ],
            Tuple[
                Union[xr.Dataset, xr.DataArray],
                DataFrame[Crosstab_df],
                DataFrame[Metrics_df],
                DataFrame[AttributeTrackingDf],
            ],
        ]
    ]:
        """
        Computes comparison between two categorical value xarray's.

        Conducts the following steps:
            - homogenize: aligns data types, spatial alignment, and rasterizes data
            - compute_agreement: computes agreement map
            - compute_crosstab: computes crosstabulation
            - compute_metrics: computes metrics

        Spatially aligning the xarray's produces copies of the original candidate and benchmark maps. To reduce memory usage, consider using the `homogenize()` accessor method to overwrite the original maps in memory or saving them on disk.

        Parameters
        ----------
        benchmark_map: Union[gpd.GeoDataFrame, xr.Dataset, xr.DataArray]
            Benchmark map.
        positive_categories: Union[Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]], default = None
            Number or list of numbers representing the values to consider as the positive condition. When the average argument is either "macro" or "weighted", this represents the categories to compute metrics for.
        comparison_function : Union[Callable, nb.np.ufunc.dufunc.DUFunc, np.ufunc, np.vectorize, str], default = 'szudzik'
            Comparison function. Created by decorating function with @nb.vectorize() or using np.ufunc(). Use of numba is preferred as it is faster. Strings with registered comparison_functions are also accepted. Possible options include "pairing_dict". If passing "pairing_dict" value, please see the description for the argument for more information on behaviour.
            All available comparison functions can be found with gval.Comparison.available_functions().
        metrics: Union[str, Iterable[str]], default = "all"
            Statistics to return in metric table.  All returns every default and registered metric.  This can be seen with gval.CatStats.available_functions().
        target_map: Optional[Union[xr.Dataset, str]], default = "benchmark"
            xarray object to match the CRS's and coordinates of candidates and benchmarks to or str with 'candidate' or 'benchmark' as accepted values.
        resampling : rasterio.enums.Resampling
            See :func:`rasterio.warp.reproject` for more details.
        pairing_dict: Optional[Dict[Tuple[Number, Number], Number]], default = None
            When "pairing_dict" is used for the comparison_function argument, a pairing dictionary can be passed by user. A pairing dictionary is structured as `{(c, b) : a}` where `(c, b)` is a tuple of the candidate and benchmark value pairing, respectively, and `a` is the value for the agreement array to be used for this pairing.

            If None is passed for pairing_dict, the allow_candidate_values and allow_benchmark_values arguments are required. For this case, the pairings in these two iterables will be paired in the order provided and an agreement value will be assigned to each pairing starting with 0 and ending with the number of possible pairings.

            A pairing dictionary can be used by the user to note which values to allow and which to ignore for comparisons. It can also be used to decide how nans are handled for cases where either the candidate and benchmark maps have nans or both.
        allow_candidate_values : Optional[Iterable[Union[int,float]]], default = None
            List of values in candidate to include in computation of agreement map. Remaining values are excluded. If "pairing_dict" is provided for comparison_function and pairing_function is None, this argument is necessary to construct the dictionary. Otherwise, this argument is optional and by default this value is set to None and all values are considered.
        allow_benchmark_values : Optional[Iterable[Union[int,float]]], default = None
            List of values in benchmark to include in computation of agreement map. Remaining values are excluded. If "pairing_dict" is provided for comparison_function and pairing_function is None, this argument is necessary to construct the dictionary. Otherwise, this argument is optional and by default this value is set to None and all values are considered.
        nodata : Optional[Number], default = None
            No data value to write to agreement map output. This will use `rxr.rio.write_nodata(nodata)`.
        encode_nodata : Optional[bool], default = False
            Encoded no data value to write to agreement map output. A nodata argument must be passed. This will use `rxr.rio.write_nodata(nodata, encode=encode_nodata)`.
        exclude_value : Optional[Number], default = None
            Value to exclude from crosstab. This could be used to denote a no data value if masking wasn't used. By default, NaNs are not cross-tabulated.
        negative_categories: Union[Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]], default=None
            Number or list of numbers representing the values to consider as the negative condition. This should be set to None when no negative categories are used or when the average type is "macro" or "weighted".
        average : str, default = "micro"
            Type of average to use when computing metrics. Options are "micro", "macro", and "weighted".
            Micro weighing computes the conditions, tp, tn, fp, and fn, for each category and then sums them.
            Macro weighing computes the metrics for each category then averages them.
            Weighted average computes the metrics for each category then averages them weighted by the number of weights argument in each category.
        weights : Optional[Iterable[Number]], default = None
            Weights to use when computing weighted average, specifically when the average argument is "weighted". Elements correspond to positive categories in order.

            Example:

            `positive_categories = [1, 2]; weights = [0.25, 0.75]`
        rasterize_attributes: Optional[list], default = None
            Numerical attributes of a Benchmark Map GeoDataFrame to rasterize.  Only applicable if benchmark map is a vector file.
            This cannot be none if the benchmark map is a vector file.
        attribute_tracking: bool, default = False
            Whether to return a dataframe with the attributes of the candidate and benchmark maps.
        attribute_tracking_kwargs: Optional[Dict], default = None
            Keyword arguments to pass to `gval.attribute_tracking()`.  This is only used if `attribute_tracking` is True. By default, agreement maps are used for attribute tracking but this can be set to None within this argument to override. See `gval.attribute_tracking` for more information.
        subsampling_df: Optional[gpd.GeoDataFrame], default = None
            DataFrame with spatial geometries and method types to subsample
        subsampling_average: Optional[str], default = None
            Way to aggregate statistics for subsamples if provided. Options are "sample", "band", and "full-detail"
            Sample calculates metrics and averages the results by subsample
            Band calculates metrics and averages all the metrics by band
            Full-detail does not aggregation on subsample or band

        Returns
        -------
        Union[
            Tuple[Union[xr.Dataset, xr.DataArray], DataFrame[Crosstab_df], DataFrame[Metrics_df]],
            Tuple[Union[xr.Dataset, xr.DataArray], DataFrame[Crosstab_df], DataFrame[Metrics_df], DataFrame[AttributeTrackingDf]]
        ]
            Tuple with agreement map/s, cross-tabulation table, and metric table. Possibly attribute tracking table as well.
        """

        # using homogenize accessor to avoid code reuse
        candidate, benchmark = self._obj.gval.homogenize(
            benchmark_map, target_map, resampling, rasterize_attributes
        )

        agreement_map = candidate.gval.compute_agreement_map(
            benchmark_map=benchmark,
            comparison_function=comparison_function,
            pairing_dict=pairing_dict,
            allow_candidate_values=allow_candidate_values,
            allow_benchmark_values=allow_benchmark_values,
            nodata=nodata,
            encode_nodata=encode_nodata,
            subsampling_df=subsampling_df,
        )

        crosstab_df = candidate.gval.compute_crosstab(
            agreement_map=agreement_map, subsampling_df=subsampling_df
        )

        metrics_df = _compute_categorical_metrics(
            crosstab_df=crosstab_df,
            metrics=metrics,
            positive_categories=positive_categories,
            negative_categories=negative_categories,
            average=average,
            weights=weights,
            sampling_average=subsampling_average,
        )

        vector_agreement = self.agreement_map_format == "vector"
        if attribute_tracking:
            results = self.__handle_attribute_tracking(
                candidate_map=candidate,
                benchmark_map=benchmark,
                agreement_map=agreement_map,
                attribute_tracking_kwargs=attribute_tracking_kwargs,
            )

            if len(results) == 2:
                attributes_df, agreement_map = results
            else:
                attributes_df = results

            agreement_map = (
                agreement_map.gval.vectorize_data()
                if vector_agreement
                else agreement_map
            )

            return agreement_map, crosstab_df, metrics_df, attributes_df

        agreement_map = (
            agreement_map.gval.vectorize_data() if vector_agreement else agreement_map
        )

        return agreement_map, crosstab_df, metrics_df

    def continuous_compare(
        self,
        benchmark_map: Union[gpd.GeoDataFrame, xr.Dataset, xr.DataArray],
        metrics: Union[str, Iterable[str]] = "all",
        target_map: Optional[Union[xr.Dataset, str]] = "benchmark",
        resampling: Optional[Resampling] = Resampling.nearest,
        nodata: Optional[Number] = None,
        encode_nodata: Optional[bool] = False,
        rasterize_attributes: Optional[list] = None,
        attribute_tracking: bool = False,
        attribute_tracking_kwargs: Optional[Dict] = None,
        subsampling_df: Optional[gpd.GeoDataFrame] = None,
        subsampling_average: str = "none",
        epsilon: float = 1e-10,
    ) -> Tuple[
        Union[
            Tuple[Union[xr.Dataset, xr.DataArray], DataFrame[Metrics_df]],
            Tuple[
                Union[xr.Dataset, xr.DataArray],
                DataFrame[Metrics_df],
                DataFrame[AttributeTrackingDf],
            ],
        ]
    ]:
        """
        Computes comparison between two continuous value xarray's.

        Conducts the following steps:
            - homogenize: aligns data types, spatial alignment, and rasterizes data
            - compute_agreement: computes agreement map which is error or candidate minus benchmark
            - compute_metrics: computes metrics

        Spatially aligning the xarray's produces copies of the original candidate and benchmark maps. To reduce memory usage, consider using the `homogenize()` accessor method to overwrite the original maps in memory or saving them on disk.

        Parameters
        ----------
        benchmark_map : Union[gpd.GeoDataFrame, xr.DataArray, xr.Dataset]
            Benchmark map.
        metrics: Union[str, Iterable[str]], default = "all"
            Statistics to return in metric table.  This can be seen with gval.ContStats.available_functions().
        target_map: Optional[Union[xr.Dataset, str]], default = "benchmark"
            xarray object to match the CRS's and coordinates of candidates and benchmarks to or str with 'candidate' or 'benchmark' as accepted values.
        resampling : rasterio.enums.Resampling
            See :func:`rasterio.warp.reproject` for more details.
        nodata : Optional[Number], default = None
            No data value to write to agreement map output. This will use `rxr.rio.write_nodata(nodata)`.
        encode_nodata : Optional[bool], default = False
            Encoded no data value to write to agreement map output. A nodata argument must be passed. This will use `rxr.rio.write_nodata(nodata, encode=encode_nodata)`.
        rasterize_attributes: Optional[list], default = None
            Numerical attributes of a GeoDataFrame to rasterize.
        attribute_tracking: bool, default = False
            Whether to return a dataframe with the attributes of the candidate and benchmark maps.
        attribute_tracking_kwargs: Optional[Dict], default = None
            Keyword arguments to pass to `gval.attribute_tracking()`.  This is only used if `attribute_tracking` is True. By default, agreement maps are used for attribute tracking but this can be set to None within this argument to override. See `gval.attribute_tracking` for more information.
        subsampling_df: Optional[gpd.GeoDataFrame], default = None
            DataFrame with spatial geometries and method types to subsample
        subsampling_average: str, default = None
            Way to aggregate statistics for subsamples if provided. Options are "sample", "band", "weighted", and "none"
            Sample calculates metrics and averages the results by subsample
            Band calculates metrics and averages all the metrics by band
            Weighted calculates metrics, scales by the weight and then averages them based on the weights
            Full-detail provides full detailed table
        epsilon : float, default = 1e-10
            Small value to avoid division by zero in some metrics.

        Returns
        -------
        Union[
            Tuple[Union[xr.Dataset, xr.DataArray], DataFrame[Metrics_df]],
            Tuple[Union[xr.Dataset, xr.DataArray], DataFrame[Metrics_df], DataFrame[AttributeTrackingDf]]
        ]
            Tuple with agreement map and metric table, possibly attribute tracking table as well.
        """

        # using homogenize accessor to avoid code reuse
        candidate, benchmark = self._obj.gval.homogenize(
            benchmark_map, target_map, resampling, rasterize_attributes
        )

        results = candidate.gval.compute_agreement_map(
            benchmark_map=benchmark,
            comparison_function=difference,
            nodata=nodata,
            encode_nodata=encode_nodata,
            subsampling_df=subsampling_df,
            continuous=True,
        )

        # If sampling_df return type gives three values assign all vars results, otherwise only agreement map results
        agreement_map, candidate_map, benchmark_map = (
            results if subsampling_df is not None else (results, candidate, benchmark)
        )

        metrics_df = _compute_continuous_metrics(
            agreement_map=agreement_map,
            candidate_map=candidate_map,
            benchmark_map=benchmark_map,
            metrics=metrics,
            subsampling_average=subsampling_average,
            subsampling_df=subsampling_df,
            epsilon=epsilon,
        )

        if attribute_tracking:
            results = self.__handle_attribute_tracking(
                candidate_map=candidate,
                benchmark_map=benchmark,
                agreement_map=agreement_map,
                attribute_tracking_kwargs=attribute_tracking_kwargs,
            )

            if len(results) == 2:
                attributes_df, agreement_map = results
            else:
                attributes_df = results

            return agreement_map, metrics_df, attributes_df

        return agreement_map, metrics_df

    def probabilistic_compare(
        self,
        benchmark_map: Union[gpd.GeoDataFrame, xr.Dataset, xr.DataArray],
        metric_kwargs: dict,
        return_on_error: Optional[Any] = None,
        target_map: Optional[Union[xr.Dataset, str]] = "benchmark",
        resampling: Optional[Resampling] = Resampling.nearest,
        rasterize_attributes: Optional[list] = None,
        attribute_tracking: bool = False,
        attribute_tracking_kwargs: Optional[Dict] = None,
    ) -> DataFrame[Prob_metrics_df]:
        """
        Computes probabilistic metrics from candidate and benchmark maps.

        Parameters
        ----------
        benchmark_map : xr.DataArray or xr.Dataset
            Benchmark map.
        metric_kwargs : dict
            Dictionary of keyword arguments to metric functions. Keys must be metrics. Values are keyword arguments to metric functions. Don't pass keys or values for 'observations' or 'forecasts' as these are handled internally with `benchmark_map` and `candidate_map`, respectively. Available keyword arguments by metric are available in `DEFAULT_METRIC_KWARGS`. If values are None or empty dictionary, default values in `DEFAULT_METRIC_KWARGS` are used.
        return_on_error : Optional[Any], default = None
            Value to return within metrics dataframe if an error occurs when computing a metric. If None, the metric is not computed and None is returned. If 'error', the raised error is returned.
        target_map: Optional[xr.Dataset or str], default = "benchmark"
            xarray object to match the CRS's and coordinates of candidates and benchmarks to or str with 'candidate' or 'benchmark' as accepted values.
        resampling : rasterio.enums.Resampling
            See :func:`rasterio.warp.reproject` for more details.
        nodata : Optional[Number], default = None
            No data value to write to agreement map output. This will use `rxr.rio.write_nodata(nodata)`.
        encode_nodata : Optional[bool], default = False
            Encoded no data value to write to agreement map output. A nodata argument must be passed. This will use `rxr.rio.write_nodata(nodata, encode=encode_nodata)`.
        rasterize_attributes: Optional[list], default = None
            Numerical attributes of a GeoDataFrame to rasterize.
        attribute_tracking: bool, default = False
            Whether to return a dataframe with the attributes of the candidate and benchmark maps.
        attribute_tracking_kwargs: Optional[Dict], default = None
            Keyword arguments to pass to `gval.attribute_tracking()`.  This is only used if `attribute_tracking` is True. By default, agreement maps are used for attribute tracking but this can be set to None within this argument to override. See `gval.attribute_tracking` for more information.

        Returns
        -------
        DataFrame[Prob_metrics_df]
            Probabilistic metrics Pandas DataFrame with computed xarray's per metric and sample.

        Raises
        ------
        ValueError
            If keyword argument is required for metric but not passed.
            If keyword argument is not available for metric but passed.
            If metric is not available.

        Warns
        -----
        UserWarning
            Warns if a metric cannot be computed. `return_on_error` determines whether the metric is not computed and None is returned or if the raised error is returned.

        References
        ----------
        .. [1] `Scoring rule <https://en.wikipedia.org/wiki/Scoring_rule#Examples_of_proper_scoring_rules>`_
        .. [2] `Strictly Proper Scoring Rules, Prediction, and Estimation <https://sites.stat.washington.edu/raftery/Research/PDF/Gneiting2007jasa.pdf>`_
        .. [3] `Properscoring/properscoring <https://github.com/properscoring/properscoring>`_
        .. [4] `xskillscore/xskillscore <https://xskillscore.readthedocs.io/en/stable/api.html#probabilistic-metrics>`_
        .. [5] `What is a proper scoring rule? <https://statisticaloddsandends.wordpress.com/2021/03/27/what-is-a-proper-scoring-rule/>`_
        """

        # agreement_map temporarily None
        agreement_map = None

        # using homogenize accessor to avoid code reuse
        candidate, benchmark = self._obj.gval.homogenize(
            benchmark_map, target_map, resampling, rasterize_attributes
        )

        metrics_df = _compute_probabilistic_metrics(
            candidate_map=candidate,
            benchmark_map=benchmark,
            metric_kwargs=metric_kwargs,
            return_on_error=return_on_error,
        )

        if attribute_tracking:
            results = self.__handle_attribute_tracking(
                candidate_map=candidate,
                benchmark_map=benchmark,
                agreement_map=agreement_map,
                attribute_tracking_kwargs=attribute_tracking_kwargs,
            )

            if len(results) == 2:
                # attributes_df, agreement_map = results
                attributes_df, _ = results
            else:
                attributes_df = results

            return agreement_map, metrics_df, attributes_df

        return agreement_map, metrics_df

    def homogenize(
        self,
        benchmark_map: Union[gpd.GeoDataFrame, xr.Dataset, xr.DataArray],
        target_map: Optional[Union[xr.Dataset, str]] = "benchmark",
        resampling: Optional[Resampling] = Resampling.nearest,
        rasterize_attributes: Optional[list] = None,
    ) -> Union[xr.Dataset, xr.DataArray]:
        """
        Homogenize candidate and benchmark maps to prepare for comparison.

        Currently supported operations include:
            - Matching projections and coordinates (spatial alignment)
            - Homogenize file formats (xarray/rasters)
            - Homogenize numerical data types (int, float, etc.).

        Parameters
        ----------
        benchmark_map: Union[gpd.GeoDataFrame, xr.Dataset, xr.DataArray]
            Benchmark map.
        target_map: Optional[Union[xr.DataArray, xr.Dataset, str]], default = "benchmark"
            xarray object to match candidates and benchmarks to or str with 'candidate' or 'benchmark' as accepted values.
        resampling: rasterio.enums.Resampling
            See :func:`rasterio.warp.reproject` for more details.
        rasterize_attributes: Optional[list], default = None
            Numerical attributes of a GeoDataFrame to rasterize

        Returns
        --------
        Union[xr.Dataset, xr.DataArray]
            Tuple with candidate and benchmark map respectively.
        """

        if isinstance(benchmark_map, gpd.GeoDataFrame):
            benchmark_map = _rasterize_data(
                candidate_map=self._obj,
                benchmark_map=benchmark_map,
                rasterize_attributes=rasterize_attributes,
            )
            self.agreement_map_format = "vector"

        self.check_same_type(benchmark_map)

        candidate, benchmark = _align_numeric_data_type(
            candidate_map=self._obj, benchmark_map=benchmark_map
        )

        candidate, benchmark = _spatial_alignment(
            candidate_map=candidate,
            benchmark_map=benchmark,
            target_map=target_map,
            resampling=resampling,
        )

        # For future operations
        candidate.gval.agreement_map_format = self.agreement_map_format

        return candidate, benchmark

    def compute_agreement_map(
        self,
        benchmark_map: Union[xr.Dataset, xr.DataArray],
        comparison_function: Union[
            Callable, nb.np.ufunc.dufunc.DUFunc, np.ufunc, np.vectorize, str
        ] = "szudzik",
        pairing_dict: Optional[Dict[Tuple[Number, Number], Number]] = None,
        allow_candidate_values: Optional[Iterable[Union[int, float]]] = None,
        allow_benchmark_values: Optional[Iterable[Union[int, float]]] = None,
        nodata: Optional[Number] = None,
        encode_nodata: Optional[bool] = False,
        subsampling_df: Optional[gpd.GeoDataFrame] = None,
        continuous: bool = False,
    ) -> Union[Union[xr.Dataset, xr.DataArray, List[Union[xr.Dataset, xr.DataArray]]]]:
        """
        Computes agreement map as xarray from candidate and benchmark xarray's.

        Parameters
        ----------
        benchmark_map : Union[xr.Dataset, xr.DataArray]
            Benchmark map.
        comparison_function : Union[Callable, nb.np.ufunc.dufunc.DUFunc, np.ufunc, np.vectorize, str], default = 'szudzik'
            Comparison function. Created by decorating function with @nb.vectorize() or using np.ufunc(). Use of numba is preferred as it is faster. Strings with registered comparison_functions are also accepted. Possible options include "pairing_dict". If passing "pairing_dict" value, please see the description for the argument for more information on behaviour.
        pairing_dict: Optional[Dict[Tuple[Number, Number], Number]], default = None
            When "pairing_dict" is used for the comparison_function argument, a pairing dictionary can be passed by user. A pairing dictionary is structured as `{(c, b) : a}` where `(c, b)` is a tuple of the candidate and benchmark value pairing, respectively, and `a` is the value for the agreement array to be used for this pairing.

            If None is passed for pairing_dict, the allow_candidate_values and allow_benchmark_values arguments are required. For this case, the pairings in these two iterables will be paired in the order provided and an agreement value will be assigned to each pairing starting with 0 and ending with the number of possible pairings.

            A pairing dictionary can be used by the user to note which values to allow and which to ignore for comparisons. It can also be used to decide how nans are handled for cases where either the candidate and benchmark maps have nans or both.
        allow_candidate_values : Optional[Iterable[Union[int,float]]], default = None
            List of values in candidate to include in computation of agreement map. Remaining values are excluded. If "pairing_dict" is set selected for comparison_function and pairing_function is None, this argument is necessary to construct the dictionary. Otherwise, this argument is optional and by default this value is set to None and all values are considered.
        allow_benchmark_values : Optional[Iterable[Union[int,float]]], default = None
            List of values in benchmark to include in computation of agreement map. Remaining values are excluded. If "pairing_dict" is set selected for comparison_function and pairing_function is None, this argument is necessary to construct the dictionary. Otherwise, this argument is optional and by default this value is set to None and all values are considered.
        nodata : Optional[Number], default = None
            No data value to write to agreement map output. This will use `rxr.rio.write_nodata(nodata)`.
        encode_nodata : Optional[bool], default = False
            Encoded no data value to write to agreement map output. A nodata argument must be passed. This will use `rxr.rio.write_nodata(nodata, encode=encode_nodata)`.
        subsampling_df : Optional[gpd.GeoDataFrame], default = None
            DataFrame with geometries to subsample data with or use as an exclusionary mask
        continuous :  bool, default = False
            Whether to return modified candidate and benchmark maps

        Returns
        -------
        Union[Union[xr.Dataset, xr.DataArray, List[Union[xr.Dataset, xr.DataArray]]]]
            Agreement map.
        """

        self.check_same_type(benchmark_map)

        agreement_range = 1 if subsampling_df is None else subsampling_df.shape[0]

        agreement_maps, candidate_maps, benchmark_maps, sample_percentage = (
            [],
            [],
            [],
            [],
        )
        for idx in range(agreement_range):
            results = (
                subsample(
                    candidate=self._obj,
                    benchmark=benchmark_map,
                    subsampling_df=subsampling_df.iloc[idx : idx + 1, :],
                )
                if subsampling_df is not None
                else [[self._obj, benchmark_map]]
            )

            candidate_copy, benchmark_copy = results[0]
            if "sample_percentage" in candidate_copy.attrs:
                sample_percentage.append(candidate_copy.attrs["sample_percentage"])

            agreement_map = Comparison.process_agreement_map(
                candidate_map=candidate_copy,
                benchmark_map=benchmark_copy,
                comparison_function=comparison_function,
                pairing_dict=pairing_dict,
                allow_candidate_values=allow_candidate_values,
                allow_benchmark_values=allow_benchmark_values,
                nodata=nodata,
                encode_nodata=encode_nodata,
                continuous=continuous,
            )

            # Preserve sampled maps for continuous statistics, otherwise delete
            if continuous:
                candidate_maps.append(candidate_copy)
                benchmark_maps.append(benchmark_copy)
            else:
                del candidate_copy, benchmark_copy

            agreement_maps.append(agreement_map)

        if subsampling_df is not None:
            if "sample_percent" in subsampling_df.columns:  # pragma: no cover
                subsampling_df.drop(columns=["sample_percent"], inplace=True)
            subsampling_df.loc[:, "sample_percent"] = sample_percentage
            return (
                (agreement_maps, candidate_maps, benchmark_maps)
                if continuous
                else agreement_maps
            )

        return agreement_maps[0]

    def compute_crosstab(
        self,
        agreement_map: Optional[
            Union[xr.DataArray, xr.Dataset, Iterable[Union[xr.DataArray, xr.Dataset]]]
        ] = None,
        subsampling_df: Optional[gpd.GeoDataFrame] = None,
    ) -> DataFrame[Crosstab_df]:
        """
        Crosstab 2 or 3-dimensional xarray DataArray to produce Crosstab DataFrame.

        Parameters
        ----------
        agreement_map : Union[xr.Dataset, xr.DataArray], default = None
            Benchmark map, {dimension}-dimensional.

        subsampling_df: Optional[gpd.GeoDataFrame], default = None
            DataFrame with spatial geometries and method types to subsample


        Returns
        -------
        DataFrame[Crosstab_df]
            Crosstab DataFrame
        """

        # Use self if agreement_map argument is not provided otherwise use agreement_map parameter
        if agreement_map is not None:
            agreement_map = (
                agreement_map if isinstance(agreement_map, list) else [agreement_map]
            )
        else:
            agreement_map = [self._obj]

        # Create cross-tabulation table for each agreement map and concatenate them
        crosstabs = []
        for idx, agreement in enumerate(agreement_map):
            crosstab = (
                _crosstab_Datasets(agreement)
                if isinstance(self._obj, xr.Dataset)
                else _crosstab_DataArrays(agreement)
            )

            if subsampling_df is not None:
                crosstab.insert(
                    0, "subsample", subsampling_df.iloc[idx]["subsample_id"]
                )

            crosstabs.append(crosstab)

        return pd.concat(crosstabs)

    def attribute_tracking(
        self,
        benchmark_map: Union[xr.DataArray, xr.Dataset],
        agreement_map: Optional[Union[xr.DataArray, xr.Dataset]] = None,
        candidate_suffix: Optional[str] = "_candidate",
        benchmark_suffix: Optional[str] = "_benchmark",
        candidate_include: Optional[Iterable[str]] = None,
        candidate_exclude: Optional[Iterable[str]] = None,
        benchmark_include: Optional[Iterable[str]] = None,
        benchmark_exclude: Optional[Iterable[str]] = None,
    ) -> Union[
        DataFrame[AttributeTrackingDf],
        Tuple[DataFrame[AttributeTrackingDf], Union[xr.DataArray, xr.Dataset]],
    ]:
        """
        Concatenate xarray attributes into a single pandas dataframe.

        Parameters
        ----------
        candidate_map : Union[xr.DataArray, xr.Dataset]
            Self. Candidate map xarray object.
        benchmark_map : Union[xr.DataArray, xr.Dataset]
            Benchmark map xarray object.
        candidate_suffix : Optional[str], default = '_candidate'
            Suffix to append to candidate map xarray attributes, by default '_candidate'.
        benchmark_suffix : Optional[str], default = '_benchmark'
            Suffix to append to benchmark map xarray attributes, by default '_benchmark'.
        candidate_include : Optional[Iterable[str]], default = None
            List of attributes to include from candidate map. candidate_include and candidate_exclude are mutually exclusive arguments.
        candidate_exclude : Optional[Iterable[str]], default = None
            List of attributes to exclude from candidate map. candidate_include and candidate_exclude are mutually exclusive arguments.
        benchmark_include : Optional[Iterable[str]], default = None
            List of attributes to include from benchmark map. benchmark_include and benchmark_exclude are mutually exclusive arguments.
        benchmark_exclude : Optional[Iterable[str]], default = None
            List of attributes to exclude from benchmark map. benchmark_include and benchmark_exclude are mutually exclusive arguments.

        Raises
        ------
        ValueError
            If candidate_include and candidate_exclude are both not None.
        ValueError
            If benchmark_include and benchmark_exclude are both not None.

        Returns
        -------
        Union[DataFrame[AttributeTrackingDf], Tuple[DataFrame[AttributeTrackingDf], Union[xr.DataArray, xr.Dataset]]]
            Pandas dataframe with concatenated attributes from candidate and benchmark maps. If agreement_map is not None, returns a tuple with the dataframe and the agreement map.
        """
        return _attribute_tracking_xarray(
            candidate_map=self._obj,
            benchmark_map=benchmark_map,
            agreement_map=agreement_map,
            candidate_suffix=candidate_suffix,
            benchmark_suffix=benchmark_suffix,
            candidate_include=candidate_include,
            candidate_exclude=candidate_exclude,
            benchmark_include=benchmark_include,
            benchmark_exclude=benchmark_exclude,
        )

    def cat_plot(
        self,
        title: str = "Categorical Map",
        colormap: str = "viridis",
        figsize: Tuple[int, int] = None,
        legend_labels: list = None,
        plot_bands: Union[str, list] = "all",
        colorbar_label: Union[str, list] = "",
        basemap: xyzservices.lib.TileProvider = cx.providers.OpenStreetMap.Mapnik,
    ):
        """
        Plots categorical Map for xarray object

        Parameters
        __________
        title : str
            Title of map, default = "Categorical Map"
        colormap : str, default = "viridis"
            Colormap of data
        figsize : tuple[int, int], default=None
            Size of the plot
        legend_labels : list, default = None
            Override labels in legend
        plot_bands: Union[str, list], default='all'
            What bands to plot
        color_bar_label : Union[str, list], default =""
            Label or labels for colorbar in the case of continuous plots
        basemap : Union[bool, xyzservices.lib.TileProvider], default = cx.providers.Stamen.Terrain
            Add basemap to the plot

        References
        ----------
        .. [1] `Matplotlib figure <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html>`_
        .. [2] `Matplotlib legend <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html>`_
        """

        return _map_plot(
            self._obj,
            title=title,
            colormap=colormap,
            figsize=figsize,
            legend_labels=legend_labels,
            plot_type="categorical",
            plot_bands=plot_bands,
            basemap=basemap,
            colorbar_label=colorbar_label,
        )

    def cont_plot(
        self,
        title: str = "Continuous Map",
        colormap: str = "viridis",
        figsize: Tuple[int, int] = None,
        plot_bands: Union[str, list] = "all",
        colorbar_label: Union[str, list] = "",
        basemap: xyzservices.lib.TileProvider = cx.providers.OpenStreetMap.Mapnik,
    ):
        """
        Plots categorical Map for xarray object

        Parameters
        __________
        title : str
            Title of map, default = "Categorical Map"
        colormap : str, default = "viridis"
            Colormap of data
        figsize : tuple[int, int], default=None
            Size of the plot
        plot_bands: Union[str, list], default='all'
            What bands to plot
        colorbar_label : Union[str, list], default =""
            Label or labels for colorbar in the case of continuous plots
        basemap : Union[bool, xyzservices.lib.TileProvider], default = cx.providers.Stamen.Terrain
            Add basemap to the plot

        References
        ----------
        .. [1] `Matplotlib figure <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html>`_
        .. [2] `Matplotlib legend <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html>`_
        """

        return _map_plot(
            self._obj,
            title=title,
            colormap=colormap,
            figsize=figsize,
            plot_type="continuous",
            plot_bands=plot_bands,
            basemap=basemap,
            colorbar_label=colorbar_label,
        )

    def vectorize_data(self) -> gpd.GeoDataFrame:
        """
        Vectorize an xarray DataArray or Dataset

        Returns
        -------
        gpd.GeoDataFrame
            Vectorized data
        """

        return _vectorize_data(self._obj)



"""
Test functionality for gval/homogenize/spatial_alignment.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from pytest_cases import parametrize_with_cases
import xarray as xr
from pytest import raises

from gval.utils.loading_datasets import _create_xarray, _create_xarray_pairs


@parametrize_with_cases(
    "upper_left, lower_right, sizes, band_params, nodata_value, encoded_nodata_value, shapes, band_dim_name, return_dataset, expected_xr",
    glob="create_xarray_*_success",
)
def test_create_xarray(
    upper_left,
    lower_right,
    sizes,
    band_params,
    nodata_value,
    encoded_nodata_value,
    shapes,
    band_dim_name,
    return_dataset,
    expected_xr,
):
    """tests creating xarray from scratch"""

    computed_xr = _create_xarray(
        upper_left,
        lower_right,
        sizes,
        band_params,
        nodata_value,
        encoded_nodata_value,
        shapes,
        band_dim_name,
        return_dataset,
    )

    xr.testing.assert_equal(computed_xr, expected_xr)


@parametrize_with_cases(
    "upper_left, lower_right, sizes, band_params_candidate, band_params_benchmark, nodata_value, encoded_nodata_value, shapes, band_dim_name, return_dataset, expected_candidate_xr, expected_benchmark_xr",
    glob="create_xarray_pairs",
)
def test_create_xarray_pairs(
    upper_left,
    lower_right,
    sizes,
    band_params_candidate,
    band_params_benchmark,
    nodata_value,
    encoded_nodata_value,
    shapes,
    band_dim_name,
    return_dataset,
    expected_candidate_xr,
    expected_benchmark_xr,
):
    """tests creating xarray from scratch"""

    computed_candidate_xr, computed_benchmark_xr = _create_xarray_pairs(
        upper_left,
        lower_right,
        sizes,
        band_params_candidate,
        band_params_benchmark,
        nodata_value,
        encoded_nodata_value,
        shapes,
        band_dim_name,
        return_dataset,
    )

    xr.testing.assert_equal(computed_candidate_xr, expected_candidate_xr)
    xr.testing.assert_equal(computed_benchmark_xr, expected_benchmark_xr)


@parametrize_with_cases(
    "upper_left, lower_right, sizes, band_params, nodata_value, encoded_nodata_value, shapes, band_dim_name, return_dataset, expected_xr",
    glob="create_xarray_unsupported_shape_fail",
)
def test_create_xarray_unsupported_shape_fail(
    upper_left,
    lower_right,
    sizes,
    band_params,
    nodata_value,
    encoded_nodata_value,
    shapes,
    band_dim_name,
    return_dataset,
    expected_xr,
):
    """tests creating xarray from scratch"""

    with raises(ValueError):
        _create_xarray(
            upper_left,
            lower_right,
            sizes,
            band_params,
            nodata_value,
            encoded_nodata_value,
            shapes,
            band_dim_name,
            return_dataset,
        )



from typing import Union, Callable
from functools import wraps
import inspect
from numbers import Number

# import numpy as np
import numba as nb
import xarray as xr

from gval.comparison.pairing_functions import (
    cantor_pair_signed,
    szudzik_pair_signed,
    difference,
    _make_pairing_dict_fn,
)
from gval.comparison.agreement import _compute_agreement_map


class ComparisonProcessing:
    """
    Class for Processing Agreement Maps and Tabulations

    Attributes
    ----------

    registered_functions : dict
        Available statistical functions with names as keys and parameters as values
    """

    def __init__(self):
        # Populates default functions for pairing functions
        self._func_names = ["pairing_dict", "cantor", "szudzik", "difference"]
        self._funcs = [
            "pairing_dict",
            cantor_pair_signed,
            szudzik_pair_signed,
            difference,
        ]

        for name, func in zip(self._func_names, self._funcs):
            setattr(self, name, func)

        self._signature_validation = {
            "names": [],
            "param_types": ["int", "float", "Number"],
            "return_type": [Number],
            "no_of_args": [2, 3],
        }

        self.registered_functions = {
            name: {"params": func_params}
            for name, func_params in zip(
                self._func_names, [["c", "b"]] * len(self._func_names)
            )
        }

    def available_functions(self) -> list:
        """
        Lists all available functions

        Returns
        -------
        List of available functions
        """
        return list(self.registered_functions.keys())

    def get_all_parameters(self):
        """
        Get all the possible arguments

        Returns
        -------
        List of all possible arguments for functions
        """

        return self._signature_validation["names"]

    def register_function(self, name: str, vectorize_func: bool = False):
        """
        Register decorator function in comparison class

        Parameters
        ----------
        name: str
            Name of function to register in comparison class
        vectorize_func: bool
            Whether to vectorize the function

        Returns
        -------
        Decorator function
        """

        def decorator(func):
            self.function_signature_check(func)

            if name not in self.registered_functions:
                self.registered_functions[name] = {
                    "params": [
                        param
                        for param in inspect.signature(func).parameters
                        if param != "self"
                    ]
                }
                # vectorize function if vectorize_func is True
                r_func = (
                    nb.vectorize(nopython=True)(func)
                    if vectorize_func is True
                    else func
                )
                setattr(self, name, r_func)
            else:
                raise KeyError("This function name already exists")

            @wraps(func)
            def wrapper(*args, **kwargs):  # pragma: no cover
                result = func(*args, **kwargs)

                return result

            return wrapper

        return decorator

    def register_function_class(self, vectorize_func: bool = False):
        """
        Register decorator function for an entire class

        Parameters
        ----------
        vectorize_func: bool
            Whether to vectorize the function

        """

        def decorator(dec_self: object):
            """
            Decorator for wrapper

            Parameters
            ----------
            dec_self: object
                Class to register pairing functions
            """

            for name, func in inspect.getmembers(dec_self, inspect.isfunction):
                if name not in self.registered_functions:
                    self.function_signature_check(func)
                    self.registered_functions[name] = {
                        "params": [
                            param
                            for param in inspect.signature(func).parameters
                            if param != "self"
                        ]
                    }
                    # vectorize funciton if vectorize_func is True
                    r_func = (
                        nb.vectorize(nopython=True)(func)
                        if vectorize_func is True
                        else func
                    )
                    setattr(self, name, r_func)
                else:
                    raise KeyError("This function name already exists")

        return decorator

    def function_signature_check(self, func):
        """
        Validates signature of registered function

        Parameters
        ----------
        func: function
            Function to check the signature of
        """
        signature = inspect.signature(func)
        names = self._signature_validation["names"]
        param_types = self._signature_validation["param_types"]
        return_type = self._signature_validation["return_type"]
        no_of_args = self._signature_validation["no_of_args"]

        # Checks if param names, type, and return type are in valid list
        # Considered no validation if either are empty
        for key, val in signature.parameters.items():
            if (key not in names and len(names) > 0) or (
                not str(val).split(": ")[-1].split(".")[-1] in param_types
                and len(param_types) > 0
            ):
                raise TypeError(
                    "Wrong parameters in function: \n"
                    f"Valid Names: {names} \n"
                    f"Valid Types: {param_types} \n"
                )

        if len(no_of_args) > 0 and len(signature.parameters) not in no_of_args:
            raise TypeError(
                "Wrong number of parameters: \n"
                f"Valid number of parameters: {no_of_args}"
            )

        if signature.return_annotation not in return_type and len(return_type) > 0:
            raise TypeError("Wrong return type \n" f"Valid return Type {return_type}")

    def get_parameters(self, func_name: str) -> list:
        """
        Get parameters of registered function

        Parameters
        ----------
        func_name: str


        Returns
        -------
        List of parameter names for the associated function
        """

        if func_name in self.registered_functions:
            return self.registered_functions[func_name]["params"]
        else:
            raise KeyError("Statistic not found in registered functions")

    def process_agreement_map(self, **kwargs) -> Union[xr.DataArray, xr.Dataset]:
        """

        Parameters
        ----------
        **kwargs

        Returns
        -------
        Union[xr.DataArray, xr.Dataset]
        Agreement map.
        """

        return self.comparison_function_from_string(func=_compute_agreement_map)(
            **kwargs
        )

    def comparison_function_from_string(
        self, func: Callable
    ) -> Callable:  # pragma: no cover
        """

        Decorator function to compose a pairing dict comparison function from a string argument

        Parameters
        ----------
        func: Callable
            Function requiring check for pairing_dict comparison function

        Returns
        -------
        Callable
            Function with appropriate comparison function
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            # NOTE: Temporary fix until better solution is found
            if "comparison_function" in kwargs and isinstance(
                kwargs["comparison_function"], str
            ):
                if kwargs["comparison_function"] in self.registered_functions:
                    kwargs["comparison_function"] = getattr(
                        self, kwargs["comparison_function"]
                    )
                else:
                    raise KeyError("Pairing function not found in registered functions")

                # In case the arguments do not exist
                kwargs["pairing_dict"] = kwargs.get("pairing_dict")
                kwargs["allow_candidate_values"] = kwargs.get("allow_candidate_values")
                kwargs["allow_benchmark_values"] = kwargs.get("allow_benchmark_values")

                if (
                    kwargs["comparison_function"] == "pairing_dict"
                ):  # when pairing_dict is a dict
                    # this creates the pairing dictionary from the passed allowed values
                    kwargs["comparison_function"] = _make_pairing_dict_fn(
                        pairing_dict=kwargs["pairing_dict"],
                        unique_candidate_values=kwargs["allow_candidate_values"],
                        unique_benchmark_values=kwargs["allow_benchmark_values"],
                    )

            if "comparison_function" not in kwargs:
                kwargs["comparison_function"] = getattr(self, "szudzik")

            # Call the decorated function
            result = func(*args, **kwargs)

            return result

        return wrapper



"""Pure Load No GVAL"""

import time
import subprocess
import os
import gc

import memray
import rioxarray as rxr


def run_normal_load(cand_file, bench_file, cache=True):
    """
    Loads without context wrappers and deletes objects
    """

    ds = rxr.open_rasterio(cand_file, cache=cache)
    ds2 = rxr.open_rasterio(bench_file, cache=cache)

    # Pure load
    ds.load()
    ds2.load()


def run_normal_load_gc(cand_file, bench_file, cache=True):
    """
    Loads without context wrappers and deletes objects
    """

    ds = rxr.open_rasterio('c_uint8.tif', cache=cache)
    ds2 = rxr.open_rasterio('b_uint8.tif', cache=cache)

    # Pure load
    ds.load()
    ds2.load()

    gc.collect()


def run_normal_load_gc(cand_file, bench_file, cache=True):
    """
    Loads without context wrappers and garbage collects
    """

    ds = rxr.open_rasterio(cand_file, cache=cache)
    ds2 = rxr.open_rasterio(bench_file, cache=cache)

    # Pure load
    ds.load()
    ds2.load()

    gc.collect()


def run_normal_load_delete(cand_file, bench_file, cache=True):
    """
    Loads without context wrappers and deletes objects
    """

    ds = rxr.open_rasterio(cand_file, cache=cache)
    ds2 = rxr.open_rasterio(bench_file, cache=cache)

    # Pure load
    ds.load()
    ds2.load()

    del ds, ds2


def run_normal_load_delete_gc(cand_file, bench_file, cache=True):
    """
    Loads without context wrappers, deletes objects, and garbage collects
    """

    ds = rxr.open_rasterio(cand_file, cache=cache)
    ds2 = rxr.open_rasterio(bench_file, cache=cache)

    # Pure load
    ds.load()
    ds2.load()

    del ds, ds2
    gc.collect()


def run_context_load(cand_file, bench_file, cache=True):
    """
    Loads with context wrappers
    """

    with (rxr.open_rasterio(cand_file, cache=cache) as ds,
          rxr.open_rasterio(bench_file, cache=cache) as ds2):

        # Pure load
        ds.load()
        ds2.load()


def run_context_load_gc(cand_file, bench_file, cache=True):
    """
    Loads with context wrappers and garbage collects
    """

    with (rxr.open_rasterio(cand_file, cache=cache) as ds,
          rxr.open_rasterio(bench_file, cache=cache) as ds2):

        # Pure load
        ds.load()
        ds2.load()

    gc.collect()


def run_context_load_delete(cand_file, bench_file, cache=True):
    """
    Loads with context wrappers and deletes objects
    """

    with (rxr.open_rasterio(cand_file, cache=cache) as ds,
          rxr.open_rasterio(bench_file, cache=cache) as ds2):

        # Pure load
        ds.load()
        ds2.load()

    del ds, ds2


def run_context_load_delete_gc(cand_file, bench_file, cache=True):
    """
    Loads with context wrappers, deletes objects, and garbage collects
    """

    with (rxr.open_rasterio(cand_file, cache=cache) as ds,
          rxr.open_rasterio(bench_file, cache=cache) as ds2):

        # Pure load
        ds.load()
        ds2.load()

    del ds, ds2
    gc.collect()


if __name__ == "__main__":

    # ---------------------- If user wants to use command line ----------------------------
    # import argparse
    #
    # # Parse arguments
    # parser = argparse.ArgumentParser(description="Flamegraph baby")
    #
    # parser.add_argument(
    #     "-o", "--file_name", help="Name of output file", required=True
    # )
    #
    # args = vars(parser.parse_args())
    # file_name = args['file_name']

    file_name = 'one81.bin'

    if os.path.exists(file_name):
        os.remove(file_name)

    with memray.Tracker(file_name):
        for x in range(5):
            run_context_load_delete_gc()

        time.sleep(2)

    subprocess.call(['memray', 'flamegraph', file_name, '--temporal'])




"""
Test functionality for computing_categorical_metrics.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"


import pandas as pd
from pytest_cases import parametrize_with_cases
from pytest import raises

from gval.comparison.compute_categorical_metrics import _compute_categorical_metrics


@parametrize_with_cases(
    "crosstab_df, metrics, positive_categories, negative_categories, expected_df",
    glob="compute_categorical_metrics_success",
)
def test_compute_categorical_metrics_success(
    crosstab_df, metrics, positive_categories, negative_categories, expected_df
):
    """tests categorical metrics functions"""

    # compute categorical metrics
    metrics_df = _compute_categorical_metrics(
        crosstab_df=crosstab_df,
        positive_categories=positive_categories,
        metrics=metrics,
        negative_categories=negative_categories,
    )

    pd.testing.assert_frame_equal(
        metrics_df, expected_df, check_dtype=False
    ), "Compute statistics did not return expected values"


@parametrize_with_cases(
    "crosstab_df, positive_categories, negative_categories, exception",
    glob="compute_categorical_metrics_fail",
)
def test_compute_categorical_metrics_fail(
    crosstab_df, positive_categories, negative_categories, exception
):
    """tests categorical metrics functions"""

    with raises(exception):
        # compute categorical metrics
        _compute_categorical_metrics(
            crosstab_df, "all", positive_categories, negative_categories
        )


@parametrize_with_cases(
    "crosstab_df, metrics, positive_categories, negative_categories, average, weights, expected_df",
    glob="compute_multi_categorical_metrics_success",
)
def test_compute_multi_categorical_metrics_success(
    crosstab_df,
    metrics,
    positive_categories,
    negative_categories,
    average,
    weights,
    expected_df,
):
    """tests multiclass categorical metrics functions"""

    # compute categorical metrics
    metrics_df = _compute_categorical_metrics(
        crosstab_df=crosstab_df,
        positive_categories=positive_categories,
        metrics=metrics,
        negative_categories=negative_categories,
        average=average,
        weights=weights,
    )

    pd.testing.assert_frame_equal(
        metrics_df, expected_df, check_dtype=False
    ), "Compute statistics did not return expected values"


@parametrize_with_cases(
    "crosstab_df, metrics, positive_categories, negative_categories, average, weights, exception",
    glob="compute_multi_categorical_metrics_fail",
)
def test_compute_multi_categorical_metrics_fail(
    crosstab_df,
    metrics,
    positive_categories,
    negative_categories,
    average,
    weights,
    exception,
):
    """tests multiclass categorical metrics functions"""

    with raises(exception):
        # compute categorical metrics
        _compute_categorical_metrics(
            crosstab_df=crosstab_df,
            positive_categories=positive_categories,
            metrics=metrics,
            negative_categories=negative_categories,
            average=average,
            weights=weights,
        )



"""
Comparison functionality
 - Includes pairing and comparison functions
 - crosstabbing functionality
 - Would np.meshgrid lend itself to pairing problem?
TODO:
    - Have not tested parallel case.
    - How to handle xr.Datasets, multiple bands, and multiple variables.
    - consider making a function registry to store pairing functions and their names in a dictionary
        - [Guide to function registration in Python with decorators](https://blog.miguelgrinberg.com/post/the-ultimate-guide-to-python-decorators-part-i-function-registration)
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Union, Iterable
from numbers import Number

import numpy as np
import pandas as pd
import xarray as xr
import pandera.pandas as pa
from pandera.typing import DataFrame
import dask
from flox.xarray import xarray_reduce

from gval.utils.schemas import Crosstab_df
from gval.utils.loading_datasets import _check_dask_array


def _crosstab_docstring(dimension: Union[int, str], xarray_obj: str = "xr.DataArray"):
    """
    Docstring decorator for crosstab functions.

    Parameters
    ----------
    dimension : Union[int, str]
        Number of dimensions function will support. Use either 2, 3, or 2/3.
    xarray_obj : str, default = "xr.DataArray"
        Type of xarray object function accepts. xr.DataArray or xr.Dataset.

    Returns
    -------
    Callable
        Decorated crosstab_* function with new docstring.

    Raises
    ------
    ValueError
        "Pass 2, 3, or 2/3 for dimension argument."
    """

    def decorator(func):
        docstring = f"""
            Crosstab {dimension}-dimensional {xarray_obj} to produce Crosstab DataFrame.

            Parameters
            ----------
            agreement_map : {xarray_obj}
                Agreement map, {dimension}-dimensional.

            Returns
            -------
            DataFrame[Crosstab_df]
                Crosstab DataFrame.

            References
            ----------
            .. [1] `flox.xarray.xarray_reduce() <https://flox.readthedocs.io/en/latest/generated/flox.xarray.xarray_reduce.html>`_
            .. [2] `xarray.rio._check_dimensions() <(https://github.com/corteva/rioxarray/blob/9d5975624fa93b76c451457a97b342ba37dfc792/rioxarray/rioxarray.py)>`_
            .. [3] `xr.rio._obj.dims <https://github.com/corteva/rioxarray/blob/9d5975624fa93b76c451457a97b342ba37dfc792/rioxarray/raster_array.py)`>_
            """
        func.__doc__ = docstring.format(dimension, xarray_obj)
        return func

    return decorator


@pa.check_types
@_crosstab_docstring(2, "xr.DataArray")
def _crosstab_2d_DataArrays(
    agreement_map: xr.DataArray,
    band_name: str = "band",
    band_value: Union[str, Number] = 1,
) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    is_dsk = False
    if _check_dask_array(agreement_map):
        agreement_map = agreement_map.drop_vars("spatial_ref")
        is_dsk = True

    agreement_map.name = "group"
    ag_dtype = agreement_map.dtype

    if is_dsk:
        agreement_counts = xarray_reduce(
            agreement_map,
            agreement_map,
            engine="numba",
            expected_groups=dask.array.unique(agreement_map.data),
            func="count",
        )
    else:
        agreement_counts = xarray_reduce(
            agreement_map, agreement_map, engine="numba", func="count"
        )

    def not_nan(number):
        return not np.isnan(number)

    # Handle pairing dictionary attribute
    pairing_dict = agreement_map.attrs["pairing_dictionary"]

    rev_dict = {}
    for k, v in pairing_dict.items():
        if np.isnan(v):
            continue
        if v in rev_dict:
            rev_dict[v].append(list(k))
        else:
            rev_dict[v] = [list(k)]

    # reorganize df to follow contingency table schema instead of xarray-spatial conventions
    crosstab_df = pd.DataFrame(
        {
            "candidate_values": [
                [y[0] for y in rev_dict[x]]
                for x in filter(not_nan, agreement_counts.coords["group"].values)
            ],
            "benchmark_values": [
                [y[1] for y in rev_dict[x]]
                for x in filter(not_nan, agreement_counts.coords["group"].values)
            ],
            "agreement_values": list(
                filter(
                    not_nan, agreement_counts.coords["group"].values.astype(ag_dtype)
                )
            ),
            "counts": [
                x
                for x, y in zip(
                    agreement_counts.values.astype(ag_dtype),
                    agreement_counts.coords["group"].values.astype(ag_dtype),
                )
                if not np.isnan(y)
            ],
        }
    )

    # Add all entries that don't exist in crosstab that exist in pairing dictionary with 0 count
    for k, v in agreement_map.attrs["pairing_dictionary"].items():
        if v not in crosstab_df["agreement_values"].values and not np.isnan(v):
            crosstab_df.loc[-1] = [k[0], k[1], v, 0]  # adding a row
            crosstab_df.index = crosstab_df.index + 1

    # Sort and reindex
    crosstab_df.sort_values(["agreement_values"], inplace=True)
    crosstab_df.reset_index()

    def is_iterable(x):
        return x[0] if isinstance(x, Iterable) else x

    # TODO  Resolve the case of multiple candidate/benchmark pairs being mapped to the same agreement
    #  value, for now just take the first pair
    crosstab_df.loc[:, "candidate_values"] = crosstab_df["candidate_values"].apply(
        is_iterable
    )
    crosstab_df.loc[:, "benchmark_values"] = crosstab_df["benchmark_values"].apply(
        is_iterable
    )

    crosstab_df.insert(0, band_name, band_value)

    return crosstab_df


@pa.check_types
@_crosstab_docstring(3, "xr.DataArray")
def _crosstab_3d_DataArrays(agreement_map: xr.DataArray) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    """
    NOTE:
    - Use of `xr.rio._obj.dims` or `xr.DataArray.rio._check_dimensions()` to get band name (sometimes called band)
        - is it always called band? may not always be band?
        - are there any other methods for doing this??
        - should this go elsewhere as it might be repeated?
        - Is this necessary? Can we remove? Under what circumstance would it not be "band"
    - xr.DataArray.rio._check_dimensions() only supports 2/3D arrays and throws this error:
        - *** rioxarray.exceptions.TooManyDimensions: Only 2D and 3D data arrays supported.
        - This is ok and is useful to restrict dimensionality.
    """
    # get band name
    band_name_agreement = agreement_map.rio._check_dimensions()

    # get coordinates for bands
    agreement_map_band_coordinates = agreement_map[band_name_agreement].values

    # cycle through extra dim
    previous_crosstab_df = None  # initializing to avoid having unset
    for i, b in enumerate(agreement_map_band_coordinates):
        crosstab_df = _crosstab_2d_DataArrays(
            agreement_map=agreement_map.sel({band_name_agreement: b}),
            band_name=band_name_agreement,
            band_value=b,
        )

        # concats crosstab_dfs across bands
        if i > 0:
            crosstab_df = pd.concat(
                [previous_crosstab_df, crosstab_df], ignore_index=True
            )

        # save df for previous band
        previous_crosstab_df = crosstab_df

    return crosstab_df


@pa.check_types
@_crosstab_docstring("2/3", "xr.DataArray")
def _crosstab_DataArrays(agreement_map: xr.DataArray) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    # TODO: these can be predicates and optional exception raising
    # 3d
    if len(agreement_map.shape) == 3:
        crosstab_func = _crosstab_3d_DataArrays
    # 2d
    elif len(agreement_map.shape) == 2:
        crosstab_func = _crosstab_2d_DataArrays
    else:
        raise ValueError(
            "Candidate and benchmark must be both 2 or 3 dimensional only."
        )

    return crosstab_func(agreement_map=agreement_map)


@pa.check_types
@_crosstab_docstring("3", "xr.Dataset")
def _crosstab_Datasets(agreement_map: xr.DataArray) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    # gets variable names
    agreement_variable_names = list(agreement_map.data_vars)

    # loop variables
    previous_crosstab_df = None  # initializing to avoid having unset
    for i, b in enumerate(agreement_variable_names):
        # Pass pairing dictionary to variable if necessary
        if (
            agreement_map[b].attrs.get("pairing_dictionary") is None
            and agreement_map.attrs.get("pairing_dictionary") is not None
        ):
            agreement_map[b].attrs["pairing_dictionary"] = agreement_map.attrs[
                "pairing_dictionary"
            ]

        crosstab_df = _crosstab_2d_DataArrays(
            agreement_map=agreement_map[b], band_value=b
        )

        # concats crosstab_dfs across bands
        if i > 0:
            crosstab_df = pd.concat(
                [previous_crosstab_df, crosstab_df], ignore_index=True
            )

        # save df for previous band
        previous_crosstab_df = crosstab_df

    return crosstab_df
